from __future__ import annotations

import logging
from datetime import datetime
from typing import Optional, Dict, List, Union, BinaryIO, Any, Tuple

from vectice import Vectice
from vectice.api import BadReferenceError, Reference, MissingReferenceError
from vectice.api.json import (
    JobArtifactType,
    ModelVersionStatus,
    RunStatus,
    JobType,
    VersionStrategy,
    ModelType,
    FileMetadata,
    StageStatus,
)
from vectice.integrations import Integrations
from vectice.models import ArtifactReference
from vectice.models import CodeVersion
from vectice.models import DatasetVersion, Dataset
from vectice.models import ModelVersion, Model
from vectice.models import Run


def _ensure_reference(item: Union[ArtifactReference, CodeVersion, ModelVersion, DatasetVersion]) -> ArtifactReference:
    if isinstance(item, ArtifactReference):
        return item
    elif isinstance(item, CodeVersion):
        return ArtifactReference(code=item.code_id, version_id=item.id)
    elif isinstance(item, ModelVersion):
        return ArtifactReference(model=item.model.id, version_id=item.id, description=item.description)
    elif isinstance(item, DatasetVersion):
        return ArtifactReference(dataset=item.dataset.id, version_id=item.id, description=item.description)
    else:
        raise RuntimeError(f"can not create artifact reference from unknown object '{item}'")


class Experiment:
    """
    High level API that manages adapter and low level client. Multiple models can belong to an Experiment.

    Manages a run and implicitly manages the dataset version, model version, code version etc of the run.

    The workflow follows: `tasks -> experiment.start() -> tasks -> experiment.complete() -> auxiliary tasks`

    As a consequence, we got by default:

    - Any Model/Dataset/Code **before** the call to :func:`~Experiment.start` is an input
    - Any Model/Dataset/Code **after** the call to :func:`~Experiment.start` is an output.

    Any Model/Dataset/Code can be assigned as an input or output if the use cases requires something beyond
    the generic workflow.

    """

    def __init__(
        self,
        job: Reference,
        project: Optional[Reference] = None,
        workspace: Optional[Reference] = None,
        user_token: Optional[str] = None,
        api_endpoint: Optional[str] = None,
        lib: Optional[object] = None,
        auto_code: bool = False,
        allow_self_signed_certificate: bool = True,
        auto_log: bool = False,
        job_type: JobType = JobType.OTHER,
        check_remote_repository: bool = True,
    ):
        """
        :param user_token: The user token to access a Vectice organisation.
        :param workspace: The workspace name or id the user wants to access.
        :param project: The project name or id the user wants to access.
        :param job: The job the user wants to work with or create.
        :param job_type: The job type the user defines.
        """
        self.job = job
        self._vectice = Vectice(
            workspace,
            project,
            user_token,
            api_endpoint,
            lib,
            allow_self_signed_certificate,
            auto_log,
        )
        self._logger = logging.getLogger(self.__class__.__name__)
        self._auto_code = auto_code
        self._check_remote_repository = check_remote_repository
        self._inputs: List[ArtifactReference] = []
        self._outputs: List[ArtifactReference] = []
        self._active_run: Optional[Run] = None
        try:
            self._job = self._vectice.get_job(job)
        except BadReferenceError:
            if isinstance(job, str):
                self._job = self._vectice.create_job(name=job, type=job_type)

    def __repr__(self):
        lib = self._vectice._integration_client
        return (
            "Experiment("
            + f"job={self._job.name}"
            + f", project={self._job.project.name}"
            + f", workspace={self._job.project.workspace.name}"
            + f", auto_code={self._auto_code}"
            + f", api_endpoint={self._vectice._client._auth.api_base_url}"
            + f", lib={lib.lib_name()}"
            if lib is not None
            else "" + ")"
        )

    @property
    def vectice(self) -> Vectice:
        return self._vectice

    @property
    def inputs(self) -> List[ArtifactReference]:
        return self._inputs.copy()

    @property
    def outputs(self) -> List[ArtifactReference]:
        return self._outputs.copy()

    def _get_run(
        self,
        create_run: bool = False,
        auto_code: Optional[bool] = None,
        check_remote_repository: Optional[bool] = None,
        run_notes: Optional[str] = None,
        run_properties: Optional[Dict] = None,
        system_name: Optional[str] = None,
    ) -> Optional[Run]:
        if self._active_run is None and create_run:
            run = self._job.create_run(
                name=f"Run {datetime.now().isoformat()}",
                properties=run_properties,
                auto_code=auto_code if auto_code is not None else self._auto_code,
                check_remote_repository=check_remote_repository
                if check_remote_repository is not None
                else self._check_remote_repository,
                notes=run_notes,
                system_name=system_name,
            )
            self._active_run = run
        return self._active_run

    def _add_artifact_reference(
        self, artifact_reference: ArtifactReference, artifact_type: Optional[JobArtifactType] = None
    ):
        run = self._get_run(create_run=False)
        if run is None:
            if self._is_input(artifact_type):
                if artifact_reference not in self._inputs:
                    self._inputs.append(artifact_reference)
            else:
                if artifact_reference not in self._outputs:
                    self._outputs.append(artifact_reference)
        else:
            if self._is_input(artifact_type):
                run.add_input(artifact_reference)
            else:
                run.add_output(artifact_reference)

    def update_run(
        self,
        run: Reference,
        job: Optional[Reference] = None,
        name: Optional[str] = None,
        system_name: Optional[str] = None,
        status: Optional[RunStatus] = None,
    ) -> Run:
        """
        Update a run.

        Updates the specified run with the new information.

        :param job: The job name or id
        :param run: The run name or id
        :param name: The name of the job
        :param system_name: The name of the system
        :param status: The status of the run

        :return: the updated Run
        """
        return self._vectice.update_run(run=run, job=job, name=name, system_name=system_name, status=status)

    def add_dataset_version(
        self,
        dataset: Reference,
        version_name: Optional[str] = None,
        version_number: Optional[int] = None,
        version_id: Optional[int] = None,
        version_strategy: VersionStrategy = VersionStrategy.AUTOMATIC,
        description: Optional[str] = None,
        is_starred: Optional[bool] = None,
        properties: Optional[Dict[str, Any]] = None,
        resources: Optional[List[str]] = None,
        metadata: Optional[List[FileMetadata]] = None,
        attachments: Optional[List[str]] = None,
        artifact_type: Optional[JobArtifactType] = None,
    ) -> ArtifactReference:
        """
        Adds a Dataset Version to the managed run.

        If the Dataset Version exists already then the existing Dataset Version is used. If the Dataset Version
        does not exist then the Dataset Version will be created and added to the managed run. The Dataset Version
        can be passed as an input or output; if it isn't then the workflow process is followed and added as
        an input or output. Resources are cloud provider assets e.g Google Cloud Storage with a connection in the
        Vectice UI and metadata are assets with no connection in Vectice UI.

        :param dataset: The dataset name or id
        :param version_name: The name of the dataset version e.g Changes 'Version 1' to 'name'
        :param description: The description of the dataset version
        :param is_starred: Whether the dataset version is starred
        :param properties: The properties of the dataset version
        :param resources: The resource uri/s of the dataset version
        :param metadata: The metadata uri/s of the dataset version
        :param artifact_type: indicate if the dataset version is an input ou an output
        :param attachments: The attachment of the code version

        :return: ArtifactReference
        """
        logging.getLogger("Project").propagate = False
        logging.getLogger("Dataset").propagate = False
        if version_id is not None or version_name is not None or version_number is not None:
            version_strategy = VersionStrategy.MANUAL

        is_exclusive = 0
        is_exclusive += 1 if version_id is not None else 0
        is_exclusive += 1 if version_name is not None else 0
        is_exclusive += 1 if version_number is not None else 0
        is_exclusive += 1 if version_strategy is not VersionStrategy.MANUAL else 0

        if is_exclusive > 1:
            raise RuntimeError(
                "You must declare only one version strategy (automatic/manual) and one version (id,name,number)"
            )

        if version_strategy == VersionStrategy.AUTOMATIC and resources is None and metadata is None:
            result = ArtifactReference(dataset=dataset, version_strategy=version_strategy)
        else:
            if version_id is not None:
                dataset_version = self.vectice.get_dataset_version(version_id, dataset)
                result = ArtifactReference(
                    dataset=dataset, version_id=version_id, description=dataset_version.description
                )
            elif version_number is not None:
                result = ArtifactReference(dataset=dataset, version_number=version_number)
            elif version_name is not None:
                try:
                    dataset_version = self.vectice.get_dataset_version(version_name, dataset)
                except Exception:
                    dataset_version = None
                    dataset_object = self.vectice.get_dataset(dataset)
                    project_id = self._vectice.project.id if self._vectice.project is not None else None
                    workspace_id = self._vectice.workspace.id if self._vectice.workspace is not None else None
                    if resources is None and metadata is None:
                        self.vectice.create_dataset_version(
                            dataset,
                            version_name,
                            project_id,
                            workspace_id,
                            description,
                            is_starred,
                            properties,
                            dataset_object.resources,
                            None,
                            attachments,
                        )
                    else:
                        self.vectice.create_dataset_version(
                            dataset,
                            version_name,
                            project_id,
                            workspace_id,
                            description,
                            is_starred,
                            properties,
                            resources,
                            metadata,
                            attachments,
                        )
                        # # self._logger.info(
                        # #     f"Changes detected automatically in resources for dataset {dataset_object.name}"
                        # # )
                        # #
                        # # self._logger.info(f"Changes detected automatically in resources for dataset {dataset_object.name}")
                        # # self._logger.info(f"Creating and using version with id: {dataset_version.id}")
                        # self._logger.info(
                        #     f"Changes detected automatically in resources for dataset {dataset_object.name}"
                        # )
                        # self._logger.info(f"Creating and using version with id: {dataset_version.id}")
                if dataset_version:
                    result = ArtifactReference(
                        dataset=dataset,
                        version_id=dataset_version.id,
                        version_name=version_name,
                        description=dataset_version.description,
                    )
                else:
                    result = ArtifactReference(dataset=dataset, version_name=version_name)
            else:
                project_id = self._vectice.project.id if self._vectice.project is not None else None
                workspace_id = self._vectice.workspace.id if self._vectice.workspace is not None else None
                dataset_object = self._vectice.get_dataset(dataset)
                if dataset_object.connection is None and resources is not None:
                    raise RuntimeError(
                        "can not create dataset version: Data resources can't be saved without a connection"
                    )
                logging.getLogger("Dataset").propagate = True
                dataset_version = self._vectice.create_dataset_version(
                    dataset,
                    None,
                    project_id,
                    workspace_id,
                    description,
                    is_starred,
                    properties,
                    resources,
                    metadata,
                    attachments,
                    version_strategy,
                )
                result = ArtifactReference(
                    dataset=dataset, version_id=dataset_version.id, description=dataset_version.description
                )
        logging.getLogger("Project").propagate = True
        logging.getLogger("Dataset").propagate = True
        self._add_artifact_reference(result, artifact_type)
        return result

    def update_dataset(
        self,
        dataset: Reference,
        name: Optional[str] = None,
        description: Optional[str] = None,
        connection: Optional[Reference] = None,
    ) -> Dataset:
        """
        Updates a dataset

        Updates a dataset. The connection is associated with the resources, e.g gcs connection for a gcs resource.

        :param dataset: dataset name or id
        :param name: The name of the dataset version e.g Changes 'Version 1' to 'name'
        :param description: The description of the dataset version
        :param connection: The connection name or id

        :return: the updated Dataset
        """
        return self._vectice.update_dataset(dataset=dataset, name=name, description=description, connection=connection)

    def update_dataset_version(
        self,
        version: Reference,
        dataset: Optional[Reference] = None,
        name: Optional[str] = None,
        description: Optional[str] = None,
        is_starred: Optional[bool] = None,
        properties: Optional[Dict[str, Any]] = None,
    ) -> DatasetVersion:
        """
        Updates the dataset version given by the user.

        :param dataset: dataset name or id
        :param version: The dataset version name or id
        :param name: The name of the dataset version e.g Changes 'Version 1' to 'name'
        :param description: The description of the dataset version
        :param is_starred: Whether the dataset version is starred
        :param properties: The properties of the dataset version

        :return: None
        """
        return self._vectice.update_dataset_version(
            version, dataset, None, None, name, description, is_starred, properties
        )

    def list_datasets(self, search: Optional[str] = None, page_index: int = 1, page_size: int = 20) -> List[Dataset]:
        """
        List all the Datasets associated with the project.

        :param search: A search term
        :param page_index: The page index
        :param page_size: The page size

        :return: A list of Datasets
        """
        return self._vectice.list_datasets(search=search, page_index=page_index, page_size=page_size)

    def list_dataset_versions(
        self, dataset: Reference, search: Optional[str] = None, page_index: int = 1, page_size: int = 20
    ) -> List[DatasetVersion]:
        """
        List the dataset versions.

        Lists the dataset versions of the specified dataset.

        :param dataset: dataset name or id

        :return: List of DatasetVersions
        """
        project_ref = None
        workspace_ref = None
        return self._vectice.list_dataset_versions(dataset, project_ref, workspace_ref, search, page_index, page_size)

    def get_dataset_version(
        self,
        version: Reference,
        dataset: Optional[Reference] = None,
    ) -> Optional[DatasetVersion]:
        """
        Get a dataset version.

        Gets a specific dataset version of the specified dataset.

        :param version: The dataset version name or id
        :param dataset: dataset name or id if the version is a name

        :return: A DatasetVersion
        """
        return self._vectice.get_dataset_version(version, dataset)

    def delete_dataset_version(
        self,
        version: Reference,
        dataset: Optional[Reference] = None,
    ) -> None:
        """
        Delete a dataset version.

        Deletes a specific dataset in the specified project.

        :param dataset: dataset name or id
        :param version: The dataset version name or id

        :return: None
        """
        self._vectice.delete_dataset_version(version, dataset)

    def start(
        self,
        auto_code: Optional[bool] = None,
        check_remote_repository: Optional[bool] = None,
        run_notes: Optional[str] = None,
        run_properties: Optional[Dict] = None,
        system_name: Optional[str] = None,
        inputs: Optional[List[Union[ArtifactReference, CodeVersion, ModelVersion, DatasetVersion]]] = None,
    ) -> Run:
        """
        Start the current experiments run.

        Once the run has started, all the Dataset/Model/Code Versions
        will be captured as outputs of the run.

        The group name allow to group several runs in one global run

        :return: Run
        """
        if inputs is not None:
            artifacts_references_inputs = [_ensure_reference(item) for item in inputs if item is not None]
            self._inputs.extend(artifacts_references_inputs)
        current_run = self._get_run(
            create_run=True,
            auto_code=auto_code,
            check_remote_repository=check_remote_repository,
            run_notes=run_notes,
            run_properties=run_properties,
            system_name=system_name,
        )
        if current_run is not None and current_run.status == RunStatus.STARTED:
            current_run.abort(reason="aborted as requesting a new Run in experiment")
            self._active_run = None
            current_run = self._get_run(
                create_run=True,
                auto_code=auto_code,
                check_remote_repository=check_remote_repository,
                run_notes=run_notes,
                run_properties=run_properties,
                system_name=system_name,
            )
        elif current_run is not None and current_run.status == RunStatus.COMPLETED:
            self._active_run = None
            self._inputs = []
            self._outputs = []
            current_run = self._get_run(
                create_run=True,
                auto_code=auto_code,
                check_remote_repository=check_remote_repository,
                run_notes=run_notes,
                run_properties=run_properties,
                system_name=system_name,
            )
        if current_run is not None:
            for input in self._inputs:
                current_run.add_input(input)
            for output in self._outputs:
                current_run.add_output(output)
            current_run.start(auto_code=auto_code, check_remote_repository=check_remote_repository)
            return current_run
        else:
            raise RuntimeError("can not create run")

    def complete(
        self, outputs: Optional[List[Union[ArtifactReference, CodeVersion, ModelVersion, DatasetVersion]]] = None
    ) -> None:
        """
        Complete the current experiment run.

        Once the run is complete. All the Dataset/Model/Code Version/s will be captured as outputs to
        the run, that were done after the experiment.start().

        :return: None
        """
        current_run = self._get_run(create_run=False)
        if current_run is not None:
            if outputs is not None:
                artifacts_references_outputs = [_ensure_reference(item) for item in outputs if item is not None]
                current_run.add_outputs(artifacts_references_outputs)
            if self._outputs is not None:
                current_run.add_outputs(self._outputs)
            current_run.complete()
            if self._vectice._auto_document_run:
                """
                Captures an existing run. That is still running.
                """
                if self._vectice._document_stage is not None:
                    try:
                        stage = self._vectice.get_stage(self._vectice._document_stage)
                    except Exception:
                        logging.debug(f"Stage {self._vectice._document_stage} for auto documentation not found.")
                        stage = self._vectice.create_stage(self._vectice._document_stage, status=StageStatus.InProgress)
                else:
                    stage = self._vectice.create_stage(
                        f"Auto Document {current_run.name}", status=StageStatus.InProgress
                    )
                current_run._capture_run_info(stage)
        self._active_run = None
        self._inputs = []
        self._outputs = []

    def fail(
        self,
        reason: Optional[str] = None,
        outputs: Optional[List[Union[ArtifactReference, CodeVersion, ModelVersion, DatasetVersion]]] = None,
    ) -> None:
        """
        Fail the current experiment run.

        Marks the experiment run as failed. Which will appear as the runs' STATUS, this can be viewed
        in the Vectice UI.

        :param reason: indicate why the run failed

        :return: None
        """
        current_run = self._get_run(create_run=False)
        if current_run is not None:
            if outputs is not None:
                artifacts_references_outputs = [_ensure_reference(item) for item in outputs if item is not None]
                current_run.add_outputs(artifacts_references_outputs)
            current_run.fail(reason=reason)
        self._active_run = None

    def abort(self, reason: Optional[str] = None) -> None:
        """
        Abort the current experiment run.

        Marks the experiment run as aborted. Which will appear as the runs' STATUS, this can be viewed
        in the Vectice UI.

        :param reason: Indicates why the run was aborted

        :return: None
        """
        current_run = self._get_run(create_run=False)
        if current_run is not None:
            current_run.abort(reason=reason)
        self._active_run = None

    def add_model_version_attachment(
        self,
        file: str,
        model_version: Optional[Reference] = None,
        model: Optional[Reference] = None,
    ) -> None:
        """
        attach a file to an existing Model Version.

        The Model Version Attachment will be added and associated to the Model Version if it exists.
        The Attachment is directly uploaded to Vectice and can be viewed immediately if the operation
        is successful. The Attachment can be seen in the Vectice UI or interacted with with the Python SDK.

        :param file: A file to attach on the model version
        :param model: The model name or id
        :param model_version: The model version name or id

        :return: None
        """
        if model is None and model_version is None:
            model, model_version = self._get_model_version_in_outputs()
            if model is None or model_version is None:
                raise MissingReferenceError("model version")
        if model_version is None:
            raise MissingReferenceError("model version")
        self._vectice.add_model_version_attachment(file, model_version, model)

    def delete_model_version_attachment(
        self,
        file: str,
        model_version: Reference,
        model: Optional[Reference] = None,
    ) -> None:
        """
        Deletes the specified attachment from the model version.

        :param model: The model name or id
        :param model_version: The model version name or id
        :param file: The file name

        :return: None
        """
        self._vectice.delete_model_version_attachment(file, model_version, model)

    def get_model_version_attachment(
        self,
        file: str,
        model_version: Reference,
        model: Optional[Reference] = None,
    ) -> Optional[BinaryIO]:
        """
        Get a model version attachment.

        Gets the attachment of a model version.

        :param model: The model name or id
        :param model_version: The model version name or id
        :param file: The file name

        :return: An Artifact Version Attachment
        """
        return self._vectice.get_model_version_attachment(file, model_version, model)

    def list_models(self, search: Optional[str] = None, page_index: int = 1, page_size: int = 20) -> List[Model]:
        """
        List all the Models associated with the project.

        :param search: A search term
        :param page_index: The page index
        :param page_size: The page size

        :return: A list of Models
        """
        return self._vectice.list_models(search=search, page_index=page_index, page_size=page_size)

    def list_model_version_attachments(
        self,
        model_version: Reference,
        model: Optional[Reference] = None,
    ) -> List[str]:
        """
        List the files attached to a Model version.

        :param model: The model name or id
        :param model_version: The model version name or id

        :return: A list of file names
        """
        return self._vectice.list_model_version_attachments(model_version, model)

    def add_dataset_version_attachment(
        self,
        file: str,
        dataset_version: Reference,
        dataset: Optional[Reference] = None,
    ) -> None:
        """
        Adds a Dataset Version Attachment to an existing Dataset Version.

        The Dataset Version Attachment will be added and associated to the Dataset Version if it exists.
        The Attachment is directly uploaded to Vectice and can be viewed immediately if the operation
        is successful. The Attachment can be seen in the Vectice UI or interacted with with the Python SDK.

        :param file: A file to attach on the dataset version
        :param dataset: The dataset name or id
        :param dataset_version: The dataset version name or id

        :return: None
        """
        self._vectice.add_dataset_version_attachment(file, dataset_version, dataset)

    def get_dataset_version_attachment(
        self,
        file: str,
        dataset_version: Reference,
        dataset: Optional[Reference] = None,
    ) -> Optional[BinaryIO]:
        """
        Get a dataset version attachment.

        Gets the attachment of a dataset version.

        :param file: The file name
        :param dataset_version: The dataset version name or id
        :param dataset: The dataset name or id

        :return: An Artifact Version Attachment
        """
        return self._vectice.get_dataset_version_attachment(file, dataset_version, dataset)

    def delete_dataset_version_attachment(
        self,
        file: str,
        dataset_version: Reference,
        dataset: Optional[Reference] = None,
    ) -> None:
        """
        Deletes the specified attachment from the dataset version.

        :param dataset: The dataset name or id
        :param dataset_version: The dataset version name or id
        :param file: The file name

        :return: None
        """
        self._vectice.delete_dataset_version_attachment(file, dataset_version, dataset)

    def list_dataset_version_attachments(
        self,
        dataset_version: Reference,
        dataset: Optional[Reference] = None,
    ) -> List[str]:
        """
        List the files attached to a Dataset version.

        :param dataset: The dataset name or id
        :param dataset_version: The dataset version name or id

        :return: An Artifact Version Attachment
        """
        return self._vectice.list_dataset_version_attachments(dataset_version, dataset)

    def add_code_version_attachment(
        self,
        file: str,
        code_version: Reference,
    ) -> None:
        """
        Adds a Code Version Attachment to an existing Code Version.

        The Code Version Attachment will be added and associated to the Code Version if it exists.
        The Attachment is directly uploaded to Vectice and can be viewed immediately if the operation
        is successful. The Attachment can be seen in the Vectice UI or interacted with with the Python SDK.


        :param file: A file to attach on the dataset version
        :param code_version: The code version name or id

        :return: None
        """
        self._vectice.add_code_version_attachment(file, code_version)

    def delete_code_version_attachment(
        self,
        file: str,
        code_version: Reference,
    ) -> None:
        """
        Delete a Code Version Attachment.

        Deletes the specified attachment from the code version.

        :param code_version: The code version name or id
        :param file: The file name

        :return: None
        """
        self._vectice.delete_code_version_attachment(file, code_version)

    def get_code_version_attachment(
        self,
        file: str,
        code_version: Reference,
    ) -> Optional[BinaryIO]:
        """
        Get a code version attachment.

        Gets the attachment of a code version.

        :param code_version: The code version name or id
        :param file: The file name

        :return: A list of file name or uri
        """
        return self._vectice.get_code_version_attachment(file, code_version)

    def list_code_version_attachments(
        self,
        code_version: Reference,
    ) -> List[str]:
        """
        List the files attached to a Code version.

        :param code_version: The code version name or id
        :return: A list of file name or uri
        """
        return self._vectice.list_code_version_attachments(code_version)

    def add_run_attachment(
        self,
        file: str,
        run: Reference,
        job: Optional[Reference] = None,
    ) -> None:
        """
        Adds a Code Version Attachment to an existing Code Version.

        The Code Version Attachment will be added and associated to the Code Version if it exists.
        The Attachment is directly uploaded to Vectice and can be viewed immediately if the operation
        is successful. The Attachment can be seen in the Vectice UI or interacted with with the Python SDK.

        :param file: A file to attach on the dataset version
        :param run: The run name or id
        :param job: The job name or id
        :return: None
        """
        return self._vectice.add_run_attachment(file, run, job)

    def delete_run_attachment(
        self,
        file: str,
        run: Reference,
        job: Optional[Reference] = None,
    ) -> None:
        """
        Delete a Code Version Attachment.

        Deletes the specified attachment from the code version.

        :param run: The run name or id
        :param job: The job name or id
        :param file: The file name

        :return: None
        """
        self._vectice.delete_run_attachment(file, run, job)

    def get_run_attachment(
        self,
        file: str,
        run: Reference,
        job: Optional[Reference] = None,
    ) -> Optional[BinaryIO]:
        """
        Get a code version attachment.

        Gets the attachment of a code version.

        :param run: The run name or id
        :param job: The job name or id
        :param file: The file name

        :return: A list of file name or uri
        """
        return self._vectice.get_run_attachment(file, run, job)

    def list_run_attachments(
        self,
        run: Reference,
        job: Optional[Reference] = None,
    ) -> List[str]:
        """
        List the files attached to a Code version.

        :param run: The run name or id
        :param job: The job name or id
        :return: A list of file name or uri
        """
        return self._vectice.list_run_attachments(run, job)

    def _is_input(self, artifact_type: Optional[JobArtifactType] = None) -> bool:
        if artifact_type == JobArtifactType.INPUT:
            return True
        elif artifact_type == JobArtifactType.OUTPUT:
            return False
        else:
            run = self._get_run(create_run=False)
            if run is None:
                return True
            else:
                return run.status == RunStatus.SCHEDULED

    def add_model_version(
        self,
        model: Reference,
        version_name: Optional[str] = None,
        version_number: Optional[int] = None,
        version_id: Optional[int] = None,
        algorithm: Optional[str] = None,
        status: ModelVersionStatus = ModelVersionStatus.EXPERIMENTATION,
        hyper_parameters: Optional[Dict] = None,
        metrics: Optional[Dict] = None,
        artifact_type: Optional[JobArtifactType] = None,
        attachment: Optional[Union[str, List[str]]] = None,
    ) -> ArtifactReference:
        """
        Adds a Model Version to the managed run.

        If the Model Version exists already then the existing Model Version is used. If the Model Version
        does not exist then the Model Version will be created and added to the managed run. The Model Version
        can be passed as an input or output; if it isn't then the workflow process is followed and added as
        an input or output.

        :param model: The model reference (model name or model id)
        :param version_name: The name of the model version (Does not assign a Version Name)
        :param algorithm: The models' algorithm
        :param status: The model status 'EXPERIMENTATION'/'STAGING'/'PRODUCTION'
        :param hyper_parameters: The hyper parameters of the model
        :param metrics: The metrics of the model
        :param artifact_type: indicate if the model version is an input ou an output
        :param attachment: The attachment of the code version

        :return: None
        """
        try:
            parent_model = self._vectice.get_model(model)
        except BadReferenceError:
            parent_model = None
        if parent_model is None and (version_id is not None or version_number is not None):
            raise ValueError("Please provide a valid model and model version reference")

        if version_id is not None and parent_model is not None:
            model_version = self._vectice.get_model_version(version=version_id, model=parent_model.id)
            result = ArtifactReference(model=model, version_id=version_id, description=model_version.description)
        elif version_name is not None and parent_model is not None:
            try:
                model_version = self._vectice.get_model_version(version=version_name, model=parent_model.id)
                result = ArtifactReference(
                    model=model,
                    version_name=version_name,
                    description=model_version.description,
                    version_id=model_version.id,
                )
            except BadReferenceError:
                result = None
        elif version_number is not None and parent_model is not None:
            result = ArtifactReference(model=model, version_number=version_number)
        else:
            result = None

        if isinstance(model, str) and parent_model is None:
            parent_model = self._vectice.create_model(model)
        elif parent_model is None:
            raise ValueError(
                f"{model} was not found and a string is needed to create a model. Please provide a valid input."
            )

        if parent_model and result is None:
            model_version = parent_model.create_model_version(
                name=version_name,
                status=status,
                algorithm_name=algorithm,
                metrics=metrics,
                hyper_parameters=hyper_parameters,
                runId=None if self._active_run is None else self._active_run.id,
            )
            result = ArtifactReference(
                model=model,
                version_name=model_version.name,
                version_id=model_version.id,
                description=model_version.description,
            )
            if attachment:
                model_version.add_attachments(attachment)
        if result:
            self._add_artifact_reference(result, artifact_type)
        else:
            raise ValueError("Please check the inputs provided.")
        return result

    def update_model(
        self,
        model: Reference,
        name: Optional[str] = None,
        description: Optional[str] = None,
        type: Optional[ModelType] = ModelType.OTHER,
    ) -> Model:
        """
        Update a model.

        :param model: The model name or id
        :param name: The name of the model
        :param description: The description of the model
        :param type: The model type

        :return: Model
        """
        return self._vectice.update_model(model=model, name=name, description=description, type=type)

    def update_model_version(
        self,
        version: Reference,
        model: Optional[Reference] = None,
        name: Optional[str] = None,
        description: Optional[str] = None,
        status: Optional[ModelVersionStatus] = ModelVersionStatus.EXPERIMENTATION,
        algorithm: Optional[str] = None,
        is_starred: Optional[bool] = False,
        metrics: Optional[Dict] = None,
        hyper_parameters: Optional[Dict] = None,
    ) -> ModelVersion:
        """
        Update a Model Version to the managed run.

        Updates the specified model version with the new params passed to this function.

        :param model: The model name or id the version belongs to
        :param version: The model version name or id
        :param name: The name of the model
        :param description: The description of the model
        :param status: The status of the model e.g 'EXPERIMENTATION'/'STAGING'/'PRODUCTION'
        :param algorithm: The algorithm the model version uses
        :param is_starred: Whether the model is starred or not
        :param metrics: The model version metrics
        :param hyper_parameters: The model version hyper parameters

        :return: None
        """
        return self._vectice.update_model_version(
            version, model, None, None, name, description, status, algorithm, is_starred, metrics, hyper_parameters
        )

    def list_model_versions(
        self, model: Reference, search: Optional[str] = None, page_index: int = 1, page_size: int = 20
    ) -> List[ModelVersion]:
        """
        List the model versions of a model.

        Lists all the versions created out of the specified model.

        :param model: The model name or id the versions belongs to

        :return: List of ModelVersions
        """
        return self._vectice.list_model_versions(model, search=search, page_index=page_index, page_size=page_size)

    def list_model_versions_dataframe(
        self, model: Reference, search: Optional[str] = None, page_index: int = 1, page_size: int = 20
    ) -> "pandas.DataFrame":  # type: ignore # noqa F821
        return self._vectice.list_model_versions_dataframe(
            model, search=search, page_index=page_index, page_size=page_size
        )

    def get_model_version(
        self,
        version: Reference,
        model: Optional[Reference] = None,
    ) -> ModelVersion:
        """
        Gets a specific version of a model.

        :param model: The model name or id the version belongs to
        :param version: The model version name or id

        :return: A ModelVersion
        """
        return self._vectice.get_model_version(version, model)

    def delete_model_version(
        self,
        version: Reference,
        model: Optional[Reference] = None,
    ) -> None:
        """
        Delete a model version.

        Deletes the specified version of a model.

        :param model: The model name or id the version belongs to
        :param version: The model version name or id

        :return: None
        """
        self._vectice.delete_model_version(version, model)

    def add_code_version(
        self,
        git_path: str = ".",
        check_remote_repository: bool = True,
        attachments: Optional[Union[str, List[str]]] = None,
    ) -> Optional[CodeVersion]:
        """
        Adds a Code Version to the managed run.

        Creates a code version in the specified project using the git repository path of the code version.

        :param git_path: The path to the git repository
        :param check_remote_repository: Indicate if we check file existence in remote repository
        :param attachments: Resources to attach to the code version e.g AWS S3 or local files

        :return: A CodeVersion
        """
        code_version = self._vectice.create_code_version(
            script_relative_path=git_path, check_remote_repository=check_remote_repository, attachments=attachments
        )
        if code_version is not None:
            self._add_artifact_reference(ArtifactReference(code=code_version.code_id, version_id=code_version.id))
        return code_version

    def add_code_version_uri(
        self,
        git_uri: str,
        entrypoint: str,
        artifact_type: Optional[JobArtifactType] = None,
        login_or_token: Optional[str] = None,
        password: Optional[str] = None,
        oauth2_token: Optional[str] = None,
        attachment: Optional[Union[str, List[str]]] = None,
    ) -> Optional[CodeVersion]:
        """
        Adds a Code Version to the managed run.

        If the Code Version exists already then the existing Code Version is used. If the Code Version
        does not exist then the Code Version will be created and added to the managed run. The Code Version
        can be passed as an input or output; if it isn't then the workflow process is followed and added as
        an input or output.

        :param git_uri: The uri of the file or folder in github/bitbucket/gitlab
        :param entrypoint: The relative path of the file or folder e.g folder/file.py
        :param artifact_type: indicate if the code version is an input ou an output
        :param login_or_token: A real login or a personal access token
        :param password: The password
        :param oauth2_token: The OAuth2 access token
        :param attachment: The attachment of the code version

        :return: None
        """

        git_version = Integrations.create_code_version_with_uri(
            git_uri, entrypoint, login_or_token, password, oauth2_token
        )
        if git_version is None:
            return None
        code_version = self._vectice.create_code_version(
            uri=git_version.uri, project=self._job.project.id, git_version=git_version
        )
        if isinstance(attachment, List):
            for item in attachment:
                if item is not None:
                    code_version.add_attachments(item)
        else:
            if attachment is not None:
                code_version.add_attachments(attachment)
        self._add_artifact_reference(
            ArtifactReference(code=code_version.code_id, version_id=code_version.id), artifact_type=artifact_type
        )
        return code_version

    def update_code_version(
        self,
        code_version: Reference,
        repository_name: str,
        branch_name: str,
        commit_hash: str,
        commit_comment: str,
        commit_author_name: str,
        commit_author_email: str,
        is_dirty: bool,
        uri: str,
        entrypoint: Optional[str] = None,
    ) -> None:
        """
        Update a code version.

        Updates the specified code version with the new given information in the params.

        :param code_version: The code version name or id
        :param repository_name: The repository name
        :param branch_name: The branch name
        :param commit_hash: The commit hash
        :param commit_comment: The commit comment
        :param commit_author_name: The commit author name
        :param commit_author_email: The commit author email
        :param is_dirty: Whether the commit is dirty
        :param uri: The uri
        :param entrypoint: The entrypoint

        :return: None
        """
        self._vectice.update_code_version(
            code_version,
            repository_name,
            branch_name,
            commit_hash=commit_hash,
            commit_comment=commit_comment,
            commit_author_name=commit_author_name,
            commit_author_email=commit_author_email,
            is_dirty=is_dirty,
            uri=uri,
            entrypoint=entrypoint,
        )

    def get_code_version(
        self,
        code_version: Reference,
    ) -> Optional[CodeVersion]:
        """
        Get a code version.

        Gets the specified version of a code artifact.

        :param code_version: The code version name or id

        :return: A CodeVersion
        """
        return self._vectice.get_code_version(code_version)

    def list_code_versions(
        self, search: Optional[str] = None, page_index: int = 1, page_size: int = 20
    ) -> Optional[List[CodeVersion]]:
        """
        List the code versions.

        Lists all the versions of a code artifact.

        :param search: A text to search for
        :param page_index: The page index
        :param page_size: The page size

        :return: A List of CodeVersions
        """
        return self._vectice.list_code_versions(search=search, page_index=page_index, page_size=page_size)

    def delete_code_version(
        self,
        code_version: Reference,
    ) -> None:
        """
        Delete a code version.

        Deletes the specified version of a code artifact.

        :param code_version: The code version name or id

        :return: None
        """
        self._vectice.delete_code_version(code_version)

    def _get_model_version_in_outputs(self) -> Tuple[Union[str, int, None], Union[str, int, None]]:
        run = self._get_run(create_run=False)
        if run:
            outputs = [item for item in run.outputs if item.model is not None]
        else:
            outputs = None
        if outputs and len(outputs) == 1:
            model = outputs[0].model
            if outputs[0].version_id:
                model_version: Union[str, int, None] = outputs[0].version_id
            elif outputs[0].version_name:
                model_version = outputs[0].version_name
            elif outputs[0].version_number:
                if model:
                    model_version = self.list_model_versions(model, search=str(outputs[0].version_number))[0].id
                else:
                    model_version = None
            else:
                model_version = None
            return model, model_version
        else:
            return None, None

    def log_metric(
        self,
        key: str,
        value: int,
        model_version: Optional[Reference] = None,
        model: Optional[Reference] = None,
        step: Optional[int] = None,
        epoch: Optional[int] = None,
    ) -> None:
        """
        Logs a metric to the Model Version specified.

        If the Model Version is not found the operation will fail. If the Model Version is found in the input or outputs
        the metrics will be directly uploaded to Vectice and assigned to the Model Version.
        A key is the metric name and the value is the metric value. Step and epoch are for deep learning models metrics.


        :param model: The model name or id
        :param model_version: The model version name or id/number
        :param key: The metric key
        :param value: The metric value
        :param step: The metric step
        :param epoch: The metric epoch

        :return: None
        """
        if model is None and model_version is None:
            model, model_version = self._get_model_version_in_outputs()
            if model is None or model_version is None:
                raise MissingReferenceError("model version")
        if model_version is None:
            raise MissingReferenceError("model version")
        self._vectice.get_model_version(model_version, model).create_metrics(metrics={key: value})

    def log_metrics(
        self,
        metrics: Dict,
        model: Optional[Reference] = None,
        model_version: Optional[Reference] = None,
        step: Optional[int] = None,
        epoch: Optional[int] = None,
    ) -> None:
        """
        Logs a dictionary of metrics to the Model Version specified.

        If the Model Version is not found the operation will fail. If the Model Version is found in the input or outputs
        the metrics will be directly uploaded to Vectice and assigned to the Model Version. A dictionary with the key,
        value, epoch and step is accepted.

        :param metrics: A dictionary of metrics
        :param model: The model name or id
        :param model_version: The model version name or id/number
        :param step: The metric step
        :param epoch: The metric epoch
        :return: None
        """
        if model is None and model_version is None:
            model, model_version = self._get_model_version_in_outputs()
            if model is None or model_version is None:
                raise MissingReferenceError("model version")
        if model_version is None:
            raise MissingReferenceError("model version")
        self._vectice.get_model_version(model_version, model).create_metrics(metrics=metrics)

    def list_metrics(
        self,
        model_version: Reference,
        model: Optional[Reference] = None,
        search: Optional[str] = None,
        page_index: int = 1,
        page_size: int = 20,
    ) -> Optional[Dict[str, Any]]:
        """
        Lists the metrics used for the specified model version.

        :param model: The model name or id the version belongs to
        :param model_version: The model version name or id
        :param search: A text to search for
        :param page_index: The page index
        :param page_size: The page size

        :return: A List of Metrics
        """
        return self._vectice.get_model_version(model_version, model).list_metrics(
            search=search, page_index=page_index, page_size=page_size
        )

    def delete_metrics(
        self,
        model_version: Reference,
        metrics: List[str],
        model: Optional[Reference] = None,
    ) -> None:
        """
        Delete a model version metric.

        Deletes a metric used for the specified model version.

        :param model: The model name or id the version belongs to
        :param model_version: The model version name or id
        :param metrics: The metric keys to delete

        :return: None
        """
        self._vectice.delete_model_version_metrics(model_version, metrics, model)

    def log_hyper_parameter(
        self,
        key: str,
        value: Union[int, float, str],
        model: Optional[Reference] = None,
        model_version: Optional[Reference] = None,
    ) -> None:
        """
        Logs a hyper parameter to the Model Version specified.

        If the Model Version is not found the operation will fail. If the Model Version is found in the input or outputs
        the hyper parameter will be directly uploaded to Vectice and assigned to the Model Version.
        A key is the hyper parameter name and the value is the metric value.

        :param key: The hyper parameter key
        :param value: The hyper parameter value
        :param model: The model name or id
        :param model_version: The model version name or id/number

        :return: None
        """
        if model is None and model_version is None:
            model, model_version = self._get_model_version_in_outputs()
            if model is None or model_version is None:
                raise MissingReferenceError("model version")
        if model_version is None:
            raise MissingReferenceError("model version")
        self._vectice.get_model_version(model_version, model).create_hyper_parameters(hyper_parameters={key: value})

    def log_hyper_parameters(
        self,
        hyper_parameters: Dict,
        model: Optional[Reference] = None,
        model_version: Optional[Reference] = None,
    ) -> None:
        """
        Update a model version hyper parameters.

        If the Model Version is not found the operation will fail. If the Model Version is found in the input or outputs
        the metrics will be directly uploaded to Vectice and assigned to the Model Version. A dictionary with multiple
        key and value pairs is accepted.

        :param hyper_parameters: The hyper_parameter ids or keys
        :param model: The model name or id the version belongs to
        :param model_version: The model version name or id

        :return: None
        """
        if model is None and model_version is None:
            model, model_version = self._get_model_version_in_outputs()
            if model is None or model_version is None:
                raise MissingReferenceError("model version")
        if model_version is None:
            raise MissingReferenceError("model version")
        self._vectice.get_model_version(model_version, model).create_hyper_parameters(hyper_parameters=hyper_parameters)

    def list_hyper_parameters(
        self,
        model_version: Reference,
        model: Optional[Reference] = None,
    ) -> Dict:
        """
        List the model version hyper parameters.

        :param model: The model name or id the version belongs to
        :param model_version: The model version name or id

        :return: A map containing the hyper parameters
        """
        return self._vectice.get_model_version(model_version, model).list_hyper_parameters()

    def delete_hyper_parameters(
        self,
        model_version: Reference,
        keys: List[str],
        model: Optional[Reference] = None,
    ) -> None:
        """
        Delete the model version hyper parameters

        Deletes the hyper parameters used for the specified model.

        :param keys: The list of keys to delete
        :param model: The model name or id the version belongs to
        :param model_version: The model version name or id

        :return: None
        """
        self._vectice.delete_model_version_hyper_parameters(model_version, keys, model)

    def document_run(self, run_id: Optional[int] = None, name: Optional[str] = None) -> None:
        """
        Captures all the assets of a run in a stage and documentation.

        :param run_id: The id of the run.
        :param name: The name of the stage.

        :return: None
        """
        if run_id is None:
            run = self._get_run(create_run=False)
        else:
            run = self._vectice.get_run(run_id)
            if run and run.status != RunStatus.COMPLETED:
                logging.warning(
                    "The run is not in a completed state and the auto documentation might not function as expected."
                )
            if name is not None:
                try:
                    stage = self._vectice.get_stage(name)
                except Exception:
                    logging.debug(f"Stage {name} for auto documentation not found.")
                    stage = self._vectice.create_stage(name, status=StageStatus.InProgress)
            else:
                stage = self._vectice.create_stage(f"Auto Document {run.name}", status=StageStatus.InProgress)
            run._capture_run_info(stage)
        if run_id is None and run is None:
            raise ValueError("Please provide a run ID or ensure there is an active run.")
        self._vectice.document_run(name=name, _experiment_run=run)



from typing import Optional, List, Union, Dict, Any

from vectice.api.json.connection_type import ConnectionType


class ParameterInput:
    def __init__(self, key: str, value: str):
        self.key = key
        self.value = value


class ConnectionInput:
    def __init__(
        self, name: str, type: ConnectionType, parameters: Optional[Dict[str, str]] = None, description: str = ""
    ):
        self.name = name
        self.parameters: Optional[List[ParameterInput]] = (
            None if parameters is None else [ParameterInput(key, value) for key, value in parameters.items()]
        )
        self.type = type
        self.description = description


class ConnectionOutput:
    def __init__(
        self,
        id: int,
        name: str,
        type: Union[ConnectionType, str],
        workspaceId: int,
        parameters: Optional[List[Dict[str, Any]]] = None,
        description: Optional[str] = None,
        *args,
        **kwargs
    ):
        self._id = int(id)
        self._name = str(name)
        self._type = type if isinstance(type, ConnectionType) else ConnectionType[str(type)]
        self._description = "" if description is None else str(description)
        self._workspaceId = int(workspaceId)
        self._parameters = (
            []
            if parameters is None
            else [ParameterInput(parameter["key"], parameter["value"]) for parameter in parameters]
        )
        self._extra = kwargs

    @property
    def id(self):
        return self._id

    @property
    def name(self):
        return self._name

    @property
    def type(self):
        return self._type

    @property
    def status(self) -> str:
        return str(self._extra["status"])

    @property
    def workspace_id(self) -> int:
        return self._workspaceId

    @property
    def description(self) -> str:
        return self._description

    @property
    def parameters(self) -> Dict[str, str]:
        return {p.key: p.value for p in self._parameters}



from __future__ import annotations
from dataclasses import dataclass
from typing import Optional
import logging
import json


@dataclass
class UserToken:
    """
    Defines the user token
    """

    name: Optional[str] = None
    """
    """
    description: Optional[str] = None
    """
    """
    key: str = ""
    """
    """

    @staticmethod
    def parse_api_token_json(token_file: str) -> UserToken:
        logging.info("Parsing api token json file...")
        with open(token_file) as file:
            data = json.load(file)
        return UserToken(**data)



from __future__ import annotations

import logging
from datetime import datetime
from typing import Optional, Union, Dict, List, TYPE_CHECKING, BinaryIO, Any

from vectice.api.json import ModelVersionStatus, PropertyInput, ArtifactVersion, UserDeclaredVersion
from vectice.api.json.metric import create_metrics_input, MetricInput
from vectice.api.json.property import create_properties_input
from vectice.models.attachment_container import AttachmentContainer
from vectice.models.metric import Metric
from vectice.models.property import Property

if TYPE_CHECKING:
    from vectice.models import Model


class ModelVersion(AttachmentContainer):
    """
    A version of the model.

    A ModelVersion store several information that can change between 2 versions like hyper parameters or metrics.

    Noteworthy:
    1. id: Backend will return id if it doesn't exist so it should be optional. SDK doesn't assign id unless we retrieve or reuse a modelVersion.
    2. model: We need the model name or id for a scenario where we won't have the modelVersionId. Currently not in the backend modelVersion.createDTO.
    3. version: Optional as we don't always need it e.g creation of model version. Currently not in the backend modelVersion.createDTO.
    4. status: Backend will throw an error if there's no status, default to EXPERIMENTATION.
    """

    def __init__(
        self,
        model: Model,
        id: int,
        name: str,
        version_number: int,
        description: Optional[str] = None,
        metrics: Optional[Dict[str, float]] = None,
        hyper_parameters: Optional[Dict[str, str]] = None,
        algorithm_name: Optional[str] = None,
        status: ModelVersionStatus = ModelVersionStatus.EXPERIMENTATION,
        is_starred: bool = False,
        user_declared_version: Optional[UserDeclaredVersion] = None,
        created_date: Optional[datetime] = None,
        updated_date: Optional[datetime] = None,
        deleted_date: Optional[datetime] = None,
    ):
        """
        :param model: The parent model of this model version
        :param id: The model version identifier
        :param name: The model version name
        :param version_number: The model version number
        :param description: The model version description
        :param metrics: The list of timestamped key/value pair
        :param hyper_parameters: List of parameters of the model
        :param algorithm_name: The algorithm name
        :param status: The status of the version
        :param is_starred: If the model isStarred
        :param user_declared_version: user declared version
        :param created_date: Creation date of the model version
        :param updated_date: Update date of the model version
        :param deleted_date: Delete date of the model version
        """
        super().__init__(name, id, model._client, "ModelVersion")
        self._model: Model = model
        self._metrics = metrics
        self._properties = hyper_parameters
        self._description = description
        self._algorithmName = algorithm_name
        self._status = status
        self._isStarred = is_starred
        self._version_number = version_number
        self._created_date = created_date
        self._updated_date = updated_date
        self._deleted_date = deleted_date
        self._user_declared_version: Optional[UserDeclaredVersion] = user_declared_version

    def __repr__(self):
        return f"ModelVersion(model={self.model}, id={self.id}, description={self.description}, metrics={self.metrics}, hyper_parameters={self.properties}, algorithm_name={self.algorithmName}, status={self.status}, is_starred={self.is_starred}, version={self.version}, user_declared_version={self.user_declared_version})"

    @property
    def is_starred(self) -> bool:
        """
        Boolean for starring the model version.
        :return: bool
        """
        return self._isStarred

    @property
    def model(self) -> Model:
        """
        The parent Model object.
        :return: Model
        """
        return self._model

    @property
    def id(self) -> int:
        """
        The model version id.
        :return: int
        """
        return self._id

    @property
    def name(self) -> str:
        """
        The model version name.
        :return: str
        """
        return self._name

    @property
    def metrics(self) -> Optional[Dict[str, float]]:
        """
        The model version metrics.
        :return: Optional[Dict[str, float]]
        """
        return self._metrics

    @property
    def properties(self) -> Optional[Dict[str, str]]:
        """
        The model version properties.
        :return: Optional[Dict[str, str]]
        """
        return self._properties

    @property
    def description(self) -> Optional[str]:
        """
        A quick description of the model version.
        :return: Optional[str]
        """
        return self._description

    @property
    def user_declared_version(self):
        """
        The user declared version.
        """
        return self._user_declared_version

    @user_declared_version.setter
    def user_declared_version(self, version: UserDeclaredVersion):
        """
        user generated version uses an empty UserDeclaredVersion()
        """
        self._user_declared_version = version

    @property
    def algorithmName(self) -> Optional[str]:
        """
        The algorithm name used for the model version.
        :return: Optional[str]
        """
        return self._algorithmName

    @property
    def status(self) -> Optional[ModelVersionStatus]:
        """
        The model version status, it can be "EXPERIMENTATION" or "STAGING" or "PRODUCTION".
        :return: Optional[ModelVersionStatus]
        """
        return self._status

    @status.setter
    def status(self, status: ModelVersionStatus):
        self._status = status

    @property
    def parent_name(self) -> Optional[str]:
        """
        The parent model name.
        :return: Optional[str]
        """
        return self._model.name

    @property
    def parent_id(self) -> Optional[int]:
        """
        The parent model id.
        :return: Optional[int]
        """
        return self._model.id

    @property
    def version(self) -> ArtifactVersion:
        """
        The model version artifact.
        :return: ArtifactVersion
        """
        return ArtifactVersion(self._version_number, self._name, self._id)

    @property
    def version_number(self) -> int:
        """
        The model version number.
        :return: int
        """
        return self._version_number

    @property
    def created_date(self) -> Optional[datetime]:
        """
        The model version number.
        :return: int
        """
        return self._created_date

    def create_metrics(self, metrics: Union[List[Metric], Dict[str, Any]]) -> None:
        """
        Create some metrics to the existing list of metrics.

        There are 2 solution to add metrics:
        - a list of already timestamped metrics.
        - a dictionary of key/value. In this case, the timestamp used is the time this method is called.

        :param metrics: The metrics e.g metric_key: value = {'metric_key': 1}

        :return: None
        """
        if isinstance(metrics, List):
            param = [MetricInput(item.key, item.value, item.timestamp) for item in metrics]
        else:
            param = create_metrics_input(metrics)
        return self._client.create_model_version_metrics(self.id, param)

    def list_metrics(
        self,
        search: Optional[str] = None,
        page_index: int = 1,
        page_size: int = 20,
    ) -> Optional[Dict[str, Any]]:
        """
        Lists the metrics used for the specified model version.

        :param search: The name to search
        :param page_index:
        :param page_size:

        :return: A List of Metrics
        """
        return {
            prop.key: prop.value
            for prop in self._client.list_model_version_metrics(
                self.id, search=search, page_size=page_size, page_index=page_index
            ).list
        }

    def delete_metrics(
        self,
        metrics: Union[List[int], List[str]],
    ) -> None:
        """
        Deletes a metric used for the specified model version.

        :param metrics: The metric ids or keys

        :return: None
        """
        metrics_reference = {
            metric.key: metric.id for metric in self._client.list_model_version_metrics(self.id).list if metric.id
        }
        for metric in metrics:
            if isinstance(metric, int):
                metric_id = metric
            elif isinstance(metric, str):
                metric_id = int(metrics_reference[metric])
            else:
                raise ValueError(f"{metric} could not be found. Please check the value.")
            try:
                self._client.delete_model_version_metrics(self.id, metric_id, self.model.id)
            except Exception as e:
                logging.warning(f"Metric {metric} failed to delete due to {e}")

    def create_hyper_parameters(self, hyper_parameters: Dict[str, Union[int, float, str]]):
        """
        Updates the hyper parameters used for the specified model version.

        :param hyper_parameters: A dict of hyper_parameter key : value

        :return: None
        """
        self._client.create_model_version_properties(self.id, create_properties_input(hyper_parameters))

    def update_hyper_parameters(self, hyper_parameters: Dict[str, Union[str, float, int]]) -> None:
        """
        Updates the hyper parameters used for the specified model version.

        :param hyper_parameters:  A dict of hyper_parameter key : value

        :return: None
        """
        properties = {prop.key: prop.id for prop in self._client.list_model_version_properties(self.id).list}
        for key, value in hyper_parameters.items():
            if properties.get(key):
                property_id = properties[key]
                property_values = Property(key, value)
                try:
                    self._client.update_model_version_properties(
                        self.id, property_id, [PropertyInput(property_values.key, property_values.value)]
                    )
                except Exception as e:
                    logging.warning(f"Hyper Parameter {key} failed to update due to {e}")

    def delete_hyper_parameters(
        self,
        hyper_parameters: Union[List[str], List[int]],
    ) -> None:
        """
        Deletes the hyper parameters used for the specified model.

        :param hyper_parameters: The hyper_parameter ids or keys

        :return: None
        """
        properties = {prop.key: prop.id for prop in self._client.list_model_version_properties(self.id).list}
        for parameter in hyper_parameters:
            if isinstance(parameter, int):
                parameter_id = parameter
            elif isinstance(parameter, str):
                parameter_id = properties[parameter]
            else:
                raise ValueError(f"{parameter} could not be found. Please check the value.")
            try:
                self._client.delete_model_version_properties(self.id, parameter_id, self.model.id)
            except Exception as e:
                logging.warning(f"Hyper Parameter {parameter} failed to delete due to {e}")

    def list_hyper_parameters(self) -> Dict[str, Any]:
        """
        List the model version hyper parameters.

        :return: A Dict of hyper_parameters
        """
        return {prop.key: prop.value for prop in self._client.list_model_version_properties(self.id).list}

    def update_metrics(
        self,
        metrics: Dict[str, int],
    ) -> None:
        """
        Update the metrics used for the specified model version

        :param metrics: The metrics e.g metric_key: value = {'metric_key': 1} OR metric_id: value = {12: 32}

        :return: None
        """
        metrics_reference = {
            metric.key: metric.id for metric in self._client.list_model_version_metrics(self.id).list if metric.id
        }
        for key, value in metrics.items():
            if metrics.get(key):
                try:
                    metric_id = metrics_reference[key]
                    metric_values = Metric(key, value)
                    self._client.update_model_version_metrics(
                        self.model.id, self.id, metric_id, [MetricInput(metric_values.key, metric_values.value)]
                    )
                except Exception as e:
                    logging.warning(f"Metric {key} failed to update due to {e}")

    def attach_model_files(self, model_object: BinaryIO) -> None:
        """
        Saves the specified model as an object to be reused.

        :param model_object: The model to save
        :return: None
        """
        attachment = [("file", (model_object.name, model_object))]
        self._client.create_attachments("modelversion", attachment, self.id)



from __future__ import annotations

from typing import Optional, List, Tuple, Union, Generic, TypeVar, TYPE_CHECKING, Callable, Sequence, Dict

import re
from urllib.parse import urlparse

if TYPE_CHECKING:
    from mypy_boto3_s3.type_defs import ObjectTypeDef
    from google.cloud.storage import Blob
    from botocore.session import Session

from vectice.api.json import FileMetadataType, FileMetadata, DataResourceSchema, SchemaColumn, DataType

NOTEBOOK = {"ipynb": True}
IMAGE_FILE = {"png": True, "jpeg": True, "svg": True}

MetadataType = TypeVar("MetadataType")


def remove_ending_slashes(path):
    # we got some case where we have several '/'
    while path.endswith("/"):
        path = path[0:-1]
    return path


class FileEntry(Generic[MetadataType]):
    def __init__(self, path: str, metadata: MetadataType):
        self.path = path
        self.metadata = metadata

    def __repr__(self):
        return f"FileEntry(path={self.path})"


def create_tree(
    entries: Sequence[FileEntry], mapper: Callable[[FileEntry], FileMetadata], protocol: str
) -> List[FileMetadata]:
    metadata_index: Dict[str, FileMetadata] = {}

    def get_entry(path: str) -> Optional[FileMetadata]:
        return metadata_index.get(path, None)

    def add_folder_entry(path: str) -> FileMetadata:
        if "/" in path:
            parent_path, name = path.rsplit("/", 1)
            parent_path = remove_ending_slashes(parent_path)
        elif path == "":
            parent_path = None
            name = ""
        else:
            parent_path = ""
            name = path

        result = FileMetadata(
            name,
            path + "/",
            parent_path + "/" if parent_path is not None else None,
            path + "/",
            FileMetadataType.Folder,
            True,
            [],
            uri=f"{protocol}://{path}/",
        )
        if parent_path is not None:
            parent = get_entry(parent_path)
            if parent is None:
                parent = add_folder_entry(parent_path)
            parent.children.append(result)
        metadata_index[path] = result
        return result

    def add_file_entry(file_entry: FileEntry):
        path = remove_ending_slashes(file_entry.path)
        parent_path = path.rsplit("/", 1)[0]
        parent = get_entry(remove_ending_slashes(parent_path))
        if parent is None:
            parent = add_folder_entry(remove_ending_slashes(parent_path))
        parent.children.append(mapper(file_entry))

    for entry in entries:
        if not entry.path.endswith("/"):
            add_file_entry(entry)

    return metadata_index[""].children


class Node(Generic[MetadataType]):
    def __init__(self, path: str, metadata: MetadataType):
        self.path = path
        self.metadata = metadata


def schema_validation(description, max_length, precision, scale):
    return (
        description if description is not None else 0,
        max_length if max_length is not None else 0,
        precision if precision is not None else 0,
        scale if scale is not None else 0,
    )


def extract_table_data(table, table_name):
    schema_columns, type_check = [], None
    for schema in table.schema:
        data_type = str(schema.field_type).lower().capitalize()
        try:
            type_check = DataType[data_type].__dict__["_value_"]
        except KeyError:
            pass
        schema_description, schema_max_length, schema_precision, schema_scale = schema_validation(
            schema.description, schema.max_length, schema.precision, schema.scale
        )
        schema_columns += [
            SchemaColumn(
                name=schema.name,
                description=schema_description,
                dataType=type_check,
                length=schema_max_length,
                precision=schema_precision,
                scale=schema_scale,
            )
        ]
    data_resource_schema = DataResourceSchema(
        type=FileMetadataType.DataTable,
        name=table_name,
        description="",
        fileFormat="bigquery#table",
        columns=schema_columns,
    )
    return data_resource_schema


def get_all_files_in_folder(client, database_name, project_id) -> Tuple[List[FileMetadata], int]:
    from google.cloud.bigquery import DatasetReference

    files_size = 0
    children: List[FileMetadata] = []
    tables = client.list_tables(database_name)

    for curr_table in tables:
        table_name = curr_table.table_id
        table_reference = DatasetReference(project_id, database_name).table(table_name)
        table = client.get_table(table_reference)
        data_resource_schema = extract_table_data(table, table_name)
        table_name, created, updated, size = table.table_id, table.created, table.modified, table.num_bytes
        files_size += size
        uri = f"bigquery://{project_id}/bigquery-public-data.{database_name}.{table_name}"
        children += [
            FileMetadata(
                name=table_name,
                uri=uri,
                itemCreatedDate=created,
                itemUpdatedDate=updated,
                size=int(files_size),
                type=FileMetadataType.DataTable,
                metadata=data_resource_schema,
            )
        ]

    return children, files_size


def extract_bigquery_metadata(
    uri: str, service_account_json_path: Optional[str] = None
) -> Optional[List[FileMetadata]]:
    """
    The terms table and dataset are switched when working with the Google BigQuery SDK and looking at the UI.
    So a table is actually a dataset in the SDK and a dataset is a table.
    """
    from google.cloud import bigquery
    from google.cloud.bigquery import DatasetReference

    matches = re.search(r"bigquery://(.*?)/(.*?)/(.+)?", uri)
    if matches:
        match_project, match_dataset, match_table = matches.group(1), matches.group(2), matches.group(3)
    elif matches is None:
        matches = re.search(r"bigquery://(.*?)/(.+)?", uri)
        match_project, match_dataset, match_table = matches.group(1), matches.group(2), None  # type: ignore
    else:
        raise ValueError("Please provide a valid project/dataset/table uri.")

    if service_account_json_path is None:
        client = bigquery.Client(project=match_project)
    else:
        client = bigquery.Client.from_service_account_json(service_account_json_path, project=match_project)
    if match_table and match_dataset:
        table_reference = DatasetReference(match_project, match_dataset).table(match_table)
        table = client.get_table(table_reference)
        data_resource_schema = extract_table_data(table, match_table)
        name, created, updated, size = table.table_id, table.created, table.modified, table.num_bytes
        uri = f"bigquery://{match_project}/bigquery-public-data.{match_dataset}.{match_table}"
        return [
            FileMetadata(
                name=name,
                uri=uri,
                itemCreatedDate=created,
                itemUpdatedDate=updated,
                size=int(size),
                type=FileMetadataType.DataTable,
                metadata=data_resource_schema,
            )
        ]
    elif match_dataset and not match_table:
        children, files_size = get_all_files_in_folder(client, match_dataset, match_project)
        uri = f"bigquery://{match_dataset}"
        return [
            FileMetadata(
                name=match_dataset,
                uri=uri,
                size=files_size,
                isFolder=True,
                children=children,
                type=FileMetadataType.Folder,
            )
        ]
    else:
        raise ValueError("Please check that a valid uri was provided.")


def decode_hash(blob):
    import base64
    import binascii

    # decode the hash provided
    base64_message = blob.md5_hash
    md5_hash = binascii.hexlify(base64.urlsafe_b64decode(base64_message))
    return md5_hash.decode("utf-8")


def get_file_type(file_name):
    file_type = file_name.split(".", 1)
    if len(file_type) == 2:
        if "csv" == file_type[1].lower():
            return "CsvFile"
        elif NOTEBOOK.get(file_type[1].lower()):
            return "Notebook"
        elif IMAGE_FILE.get(file_type[1].lower()):
            return "ImageFile"
        elif "md" == file_type[1].lower():
            return "MdFile"
        else:
            return "File"
    else:
        return "File"


def extract_gcs_metadata(
    uri: Union[str, List[str]], service_account_json_path: Optional[str] = None
) -> Optional[List[FileMetadata]]:
    def create_metadata(entry: FileEntry[Blob]) -> FileMetadata:
        parts = entry.path.rsplit("/")
        filename = parts[-1]
        return FileMetadata(
            filename,
            entry.path,
            parts[0],
            entry.path,
            get_file_type(filename),
            entry.path.endswith("/"),
            [],
            entry.metadata.size,
            f"gs://{entry.path}",
            str(entry.metadata.generation),
            None,
            decode_hash(entry.metadata),
            entry.metadata.time_created,
            entry.metadata.updated,
        )

    try:
        from google.cloud import storage  # type: ignore

        if service_account_json_path is None:
            storage_client = storage.Client()
        else:
            storage_client = storage.Client.from_service_account_json(service_account_json_path)
        if isinstance(uri, str):
            uri = [uri]

        entries = []
        for item in uri:
            parsed_uri = urlparse(item)
            bucket_name = parsed_uri.netloc
            blob = parsed_uri.path[1:] if parsed_uri.path.startswith("/") else parsed_uri.path
            blobs_query = list(storage_client.list_blobs(bucket_name, prefix=blob))
            entries.extend([FileEntry(f"{bucket_name}/{(blob.name)}", blob) for blob in blobs_query])
        return create_tree(entries, create_metadata, "gs")
    except ModuleNotFoundError:
        raise RuntimeError("The google-cloud-storage package is needed to use this functionality")


def extract_s3_metadata(
    uri: Union[str, List[str]],
    aws_access_key_id: Optional[str] = None,
    aws_secret_access_key: Optional[str] = None,
    aws_session_token: Optional[str] = None,
    profile_name: Optional[str] = None,
    region_name: Optional[str] = None,
    botocore_session: Optional[Session] = None,
) -> Optional[List[FileMetadata]]:
    """
    For credentials, please read AWS instructions at https://boto3.amazonaws.com/v1/documentation/api/latest/guide/credentials.html
    As describe the AWS client also support environment variables `AWS_ACCESS_KEY_ID `, `AWS_SECRET_ACCESS_KEY` and `AWS_SESSION_TOKEN`.
    It also support credentials file in `~/.aws/credentials`.

    :param uri: list of files/folders to be used for the dataset. the metadata will be extracted from this list.
    :type aws_access_key_id: string
    :param aws_access_key_id: AWS access key ID
    :type aws_secret_access_key: string
    :param aws_secret_access_key: AWS secret access key
    :type aws_session_token: string
    :param aws_session_token: AWS temporary session token
    :type region_name: string
    :param region_name: Default region when creating new connections
    :type profile_name: string
    :param profile_name: The name of a profile to use. If not given, then
                         the default profile is used.
    :type botocore_session: botocore.session.Session
    :param botocore_session: Use this Botocore session instead of creating
                             a new default one.
    :return:
    """

    def create_metadata(entry: FileEntry) -> FileMetadata:
        parts = entry.path.rsplit("/")
        filename = parts[-1]
        return FileMetadata(
            name=filename,
            isFolder=False,
            digest=entry.metadata["ETag"],
            path=entry.path,
            type=get_file_type(filename),
            size=entry.metadata["Size"],
            uri=f"s3://{entry.path}",
            itemUpdatedDate=entry.metadata["LastModified"],
            generation=None if "VersionId" not in entry.metadata else entry.metadata["VersionId"],
        )

    def create_file_entry(client, bucket_name, is_bucket_versioned, blob) -> FileEntry:
        if is_bucket_versioned:
            response = client.list_object_versions(Prefix=blob["Key"], Bucket=bucket_name)
            versions = [*response["Versions"]]
            current = list(filter(lambda v: v["IsLatest"] and v["Key"] == blob["Key"], versions))[0]
        else:
            current = blob
        return FileEntry(f"{bucket_name}/{current['Key']}", current)

    try:
        import boto3

        session = boto3.session.Session(
            profile_name=profile_name,
            aws_session_token=aws_session_token,
            aws_access_key_id=aws_access_key_id,
            aws_secret_access_key=aws_secret_access_key,
            region_name=region_name,
            botocore_session=botocore_session,
        )
        client = session.client("s3")
        if isinstance(uri, str):
            uri = [uri]
        entries = []
        for item in uri:
            parsed_uri = urlparse(item)
            bucket_name = parsed_uri.netloc
            response = client.get_bucket_versioning(Bucket=bucket_name)
            is_bucket_versioned = "Status" in response and response["Status"] == "Enabled"
            prefix = parsed_uri.path[1:] if parsed_uri.path.startswith("/") else parsed_uri.path
            blobs_query: List[ObjectTypeDef] = client.list_objects_v2(Bucket=bucket_name, Prefix=prefix)["Contents"]
            entries.extend([create_file_entry(client, bucket_name, is_bucket_versioned, blob) for blob in blobs_query])
        return create_tree(entries, create_metadata, "s3")
    except ModuleNotFoundError:
        raise RuntimeError("The boto3 package is needed to use this functionality")


class DatasetMetadata:
    @classmethod
    def create_bigquery(cls, uri: str, service_account_json_path: Optional[str] = None) -> Optional[List[FileMetadata]]:
        dataset_metadata_artifact = extract_bigquery_metadata(uri, service_account_json_path)
        return dataset_metadata_artifact

    @classmethod
    def create_gcs(
        cls, uri: Union[str, List[str]], service_account_json_path: Optional[str] = None
    ) -> Optional[List[FileMetadata]]:
        dataset_metadata_artifact = extract_gcs_metadata(uri, service_account_json_path)
        return dataset_metadata_artifact

    @classmethod
    def create_s3(
        cls,
        uri: Union[str, List[str]],
        aws_access_key_id: Optional[str] = None,
        aws_secret_access_key: Optional[str] = None,
        aws_session_token: Optional[str] = None,
        profile_name: Optional[str] = None,
        region_name: Optional[str] = None,
        botocore_session: Optional[Session] = None,
    ) -> Optional[List[FileMetadata]]:
        """
        For credentials, please read AWS instructions at https://boto3.amazonaws.com/v1/documentation/api/latest/guide/credentials.html
        As describe the AWS client also support environment variables `AWS_ACCESS_KEY_ID `, `AWS_SECRET_ACCESS_KEY` and `AWS_SESSION_TOKEN`.
        It also support credentials file in `~/.aws/credentials`.

        :param uri: list of files/folders to be used for the dataset. the metadata will be extracted from this list.
        :type aws_access_key_id: string
        :param aws_access_key_id: AWS access key ID
        :type aws_secret_access_key: string
        :param aws_secret_access_key: AWS secret access key
        :type aws_session_token: string
        :param aws_session_token: AWS temporary session token
        :type region_name: string
        :param region_name: Default region when creating new connections
        :type profile_name: string
        :param profile_name: The name of a profile to use. If not given, then
                             the default profile is used.
        :type botocore_session: botocore.session.Session
        :param botocore_session: Use this Botocore session instead of creating
                                 a new default one.
        :return:
        """
        dataset_metadata_artifact = extract_s3_metadata(
            uri,
            aws_access_key_id,
            aws_secret_access_key,
            aws_session_token,
            profile_name,
            region_name,
            botocore_session,
        )
        return dataset_metadata_artifact



from __future__ import annotations

import logging
from typing import List, Optional, TYPE_CHECKING, Tuple, BinaryIO, Any, Dict, Sequence

from vectice.api._auth import Auth
from vectice.api.artifact import ArtifactApi
from vectice.api.code import CodeApi
from vectice.api.code_version import CodeVersionApi
from vectice.api.connections import ConnectionApi
from vectice.api.dataset import DatasetApi
from vectice.api.dataset_version import DatasetVersionApi
from vectice.api.job import JobApi
from vectice.api.json_object import JsonObject
from vectice.api.run import RunApi
from .attachment import AttachmentApi
from .compatibility import CompatibilityApi
from .stage import StageApi
from .block import BlockApi
from .json import (
    ModelVersionOutput,
    ModelVersionInput,
    PagedResponse,
    ModelOutput,
    ModelInput,
    JobInput,
    ArtifactOutput,
    ProjectInput,
    ConnectionInput,
    Page,
    StartRunInput,
    StopRunInput,
    StopRunOutput,
    FileMetadata,
    StageInput,
    StageOutput,
    BlockInput,
    BlockOutput,
)
from .json.compatibility import CompatibilityOutput
from .json.metric import MetricInput, MetricOutput
from .json.property import PropertyOutput, PropertyInput
from .json.rule import ArtifactReferenceInput, FillRunInput
from .json.workspace import WorkspaceInput
from .model import ModelApi
from .model_version import ModelVersionApi
from .project import ProjectApi
from .reference import MissingReferenceError
from .rule import RuleApi
from .workspace import WorkspaceApi

if TYPE_CHECKING:
    from vectice import Reference
    from .json import JobOutput
    from .json import RunOutput, RunInput
    from .json import WorkspaceOutput
    from .json import ProjectOutput
    from .json import ArtifactInput
    from .json import ConnectionOutput
    from .json import DatasetInput, DatasetOutput
    from .json import DatasetVersionOutput, DatasetVersionInput
    from .json import AttachmentOutput
    from .json import CodeOutput, CodeInput
    from .json import CodeVersionOutput, CodeVersionInput


class Client:
    """
    Low level Vectice API client.
    """

    def __init__(
        self,
        workspace: Optional[Reference] = None,
        project: Optional[Reference] = None,
        token: Optional[str] = None,
        api_endpoint: Optional[str] = None,
        auto_connect=True,
        allow_self_certificate=True,
    ):
        self._auth = Auth(
            api_endpoint=api_endpoint,
            api_token=token,
            auto_connect=auto_connect,
            allow_self_certificate=allow_self_certificate,
        )
        self._logger = logging.getLogger(self.__class__.__name__)
        self._workspace = None
        self._project = None
        if auto_connect and workspace is not None:
            if isinstance(project, int):
                self._project = self.get_project(project)
                self._workspace = self._project.workspace
                if workspace is not None:
                    if isinstance(workspace, str):
                        if workspace != self._workspace.name:
                            raise ValueError(
                                f"Inconsistency in configuration: project {project} does not belong to workspace {workspace}"
                            )
                    else:
                        if workspace != self._workspace.id:
                            raise ValueError(
                                f"Inconsistency in configuration: project {project} does not belong to workspace {workspace}"
                            )
                self._logger.info(
                    f"Successfully authenticated. You'll be working on Project: {self._project.name} part of Workspace: {self._workspace.name}"
                )
            else:
                self._workspace = self.get_workspace(workspace)
                if project is not None:
                    self._project = self.get_project(project, workspace)
                    self._logger.info(
                        f"Successfully authenticated. You'll be working on Project: {self._project.name} part of Workspace: {self._workspace.name}"
                    )
                else:
                    self._logger.info(f"The entered Workspace is OK and allows you to work on {self._workspace.name}")
        elif auto_connect and workspace is None and project:
            self._project = self.get_project(project)
            if self._project is not None:
                self._workspace = self._project.workspace
            self._logger.info(
                f"Successfully authenticated. You'll be working on Project: {self._project.name} part of Workspace: {self._workspace.name}"
            )

    @property
    def workspace(self) -> Optional[WorkspaceOutput]:
        """
        The workspace object.
        :return: WorkspaceOutput
        """
        return self._workspace

    @property
    def project(self) -> Optional[ProjectOutput]:
        """
        The project object.
        :return: ProjectOutput
        """
        return self._project

    def check_compatibility(self) -> CompatibilityOutput:
        return CompatibilityApi(self._auth).check_version()

    def create_project(self, data: ProjectInput, workspace: Reference) -> ProjectOutput:
        """
         Creates a project.

        :param data: The ProjectInput json structure
        :param workspace: The workspace name or id

        :return: The ProjectOutput json structure
        """
        result = ProjectApi(self._auth).create_project(data, workspace)
        self._logger.info(f"Project with id: {result.id} successfully created.")
        return result

    def delete_project(self, project: Reference, workspace: Optional[Reference] = None):
        """
         Deletes a project.

        :param project: The project name or id
        :param workspace: The workspace name or id

        :return: None
        """
        ProjectApi(self._auth).delete_project(project, workspace)

    def update_project(self, data: ProjectInput, project: Reference, workspace: Reference) -> ProjectOutput:
        """
        Updates a project.

        :param data: The ProjectInput json structure
        :param project: The project name or id
        :param workspace: The workspace name or id

        :return: The ProjectOutput json structure
        """
        return ProjectApi(self._auth).update_project(data, project, workspace)

    def list_projects(
        self,
        workspace: Reference,
        search: Optional[str] = None,
        page_index: int = Page.index,
        page_size: int = Page.size,
    ) -> PagedResponse[ProjectOutput]:
        """
        Lists the projects in a workspace.

        :param workspace: The workspace name or id
        :param search: A text to search for
        :param page_index: The index of the page
        :param page_size: The size of the page

        :return: PagedResponse[ProjectOutput]
        """
        return ProjectApi(self._auth).list_projects(workspace, search, page_index, page_size)

    def get_project(self, project: Reference, workspace: Optional[Reference] = None) -> ProjectOutput:
        """
        Gets a project.

        :param project: The project name or id
        :param workspace: The workspace name or id

        :return: The ProjectOutput json structure
        """
        return ProjectApi(self._auth).get_project(project, workspace)

    def get_workspace(self, workspace: Reference) -> WorkspaceOutput:
        """
        Gets a workspace.

        :param workspace: The workspace name or id

        :return: The WorkspaceOutput json structure
        """
        return WorkspaceApi(self._auth).get_workspace(workspace)

    def create_workspace(self, data: WorkspaceInput) -> WorkspaceOutput:
        """
        Creates a workspace.

        :param data: The WorkspaceInput json structure

        :return: The WorkspaceOutput json structure
        """
        result = WorkspaceApi(self._auth).create_workspace(data)
        return result

    def update_workspace(self, data: WorkspaceInput, workspace: Reference) -> WorkspaceOutput:
        """
        Updates a workspace.

        :param data: The WorkspaceInput json structure
        :param workspace: The workspace name or id

        :return: The WorkspaceOutput json structure
        """
        return WorkspaceApi(self._auth).update_workspace(data, workspace)

    def list_workspaces(
        self, search: Optional[str] = None, page_index: int = 1, page_size: int = 20
    ) -> PagedResponse[WorkspaceOutput]:
        """
        Lists the workspaces.

        :param search: A text to search for
        :param page_index: The index of the page
        :param page_size: The size of the page

        :return: PagedResponse[WorkspaceOutput]
        """
        return WorkspaceApi(self._auth).list_workspaces(search, page_index, page_size)

    def get_job(
        self, job: Reference, project: Optional[Reference] = None, workspace: Optional[Reference] = None
    ) -> JobOutput:
        """
        Gets a job.

        :param job: The job name or id
        :param project: The project name or id
        :param workspace: The workspace name or id

        :return: The JobOutput json structure
        """
        if project is None and self.project is not None:
            project = self.project.id
        if workspace is None and self.workspace is not None:
            workspace = self.workspace.id
        return JobApi(self._auth).get_job(job, project, workspace)

    def delete_job(
        self, job: Reference, project: Optional[Reference] = None, workspace: Optional[Reference] = None
    ) -> None:
        """
        Deletes a job.

        :param job: The job name or id
        :param project: The project name or id
        :param workspace: The workspace name or id

        :return: None
        """

        if project is None and self.project is not None:
            project = self.project.id
        if workspace is None and self.workspace is not None:
            workspace = self.workspace.id
        JobApi(self._auth).delete_job(job, project, workspace)

    def start_run(
        self, data: StartRunInput, project: Optional[Reference] = None, workspace: Optional[Reference] = None
    ) -> JsonObject:
        """
        Start a run.

        :param data: The StartRunInput json structure
        :param project: The project name or id
        :param workspace: The workspace name or id

        :return: A json object
        """
        if project is None and self._project is not None:
            project = self._project.id
        if project is None:
            raise MissingReferenceError("run", "project")
        return RuleApi(self._auth).start_run(data, project, workspace)

    def fill_run(
        self,
        run: Reference,
        job: Optional[Reference] = None,
        project: Optional[Reference] = None,
        workspace: Optional[Reference] = None,
        inputs: Optional[List[ArtifactReferenceInput]] = None,
        outputs: Optional[List[ArtifactReferenceInput]] = None,
    ):
        """
        Fills a run.

        :param run: The run name or id
        :param job: The job name or id
        :param project: The project name or id
        :param workspace: The workspace name or id
        :param inputs: The list of artifact references
        :param outputs: The list of artifact references

        :return:  FillRunOutput
        """
        if inputs is not None or outputs is not None:
            run_object = self.get_run(run, job, project, workspace)
            return RuleApi(self._auth).fill_run(
                FillRunInput(run_object.id, [] if inputs is None else inputs, [] if outputs is None else outputs)
            )

    def stop_run(
        self,
        data: StopRunInput,
    ) -> StopRunOutput:
        """
        Stops a run.

        :param data: The StopRunInput json structure

        :return: The StopRunOutput json structure
        """
        return RuleApi(self._auth).stop_run(data)

    def list_jobs(
        self,
        project: Optional[Reference] = None,
        workspace: Optional[Reference] = None,
        search: Optional[str] = None,
        page_index=Page.index,
        page_size=Page.size,
    ) -> PagedResponse[JobOutput]:
        """
        Lists the jobs in a project.

        :param project: The project name or id
        :param workspace: The workspace name or id
        :param search: A text to filter jobs we are looking for
        :param page_index: The index of the page
        :param page_size: The size of the page

        :return: PagedResponse[JobOutput]
        """
        if project is None and self.project is not None:
            project = self.project.id
        if workspace is None and self.workspace is not None:
            workspace = self.workspace.id
        if project is None:
            raise MissingReferenceError("project")
        return JobApi(self._auth).list_jobs(project, workspace, search, page_index, page_size)

    def create_job(
        self, data: JobInput, project: Optional[Reference] = None, workspace: Optional[Reference] = None
    ) -> JobOutput:
        """
        Creates a job.

        :param data: The JobInput json structure
        :param project: The project name or id
        :param workspace: The workspace name or id

        :return: A JobOutput instance
        """
        if project is None and self.project is not None:
            project = self.project.id
        if workspace is None and self.workspace is not None:
            workspace = self.workspace.id
        if project is None:
            raise MissingReferenceError("project")
        return JobApi(self._auth).create_job(project, workspace, data)

    def update_job(self, data: JobOutput, project: Optional[Reference] = None, workspace: Optional[Reference] = None):
        """
        Updates a job.

        :param data: The JobOutput json structure
        :param project: The project name or id
        :param workspace: The workspace name or id

        :return: The JobOutput json structure
        """

        if project is None and self.project is not None:
            project = self.project.id
        if workspace is None and self.workspace is not None:
            workspace = self.workspace.id
        return JobApi(self._auth).update_job(project, workspace, data)

    def list_runs(
        self,
        job: Reference,
        project: Optional[Reference] = None,
        workspace: Optional[Reference] = None,
        search: Optional[str] = None,
        page_index=Page.index,
        page_size=Page.size,
    ) -> PagedResponse[RunOutput]:
        """
        Lists the runs of a specific job.

        :param job: The job name or id
        :param project: The project name or id
        :param workspace: The workspace name or id
        :param search: A name to search for
        :param page_index: The index of the page
        :param page_size: The size of the page

        :return: PagedResponse[RunOutput]
        """
        result = RunApi(self._auth).list_runs(job, project, workspace, search, page_index, page_size)
        return result

    def create_run(
        self,
        data: RunInput,
        job: Reference,
        project: Optional[Reference] = None,
        workspace: Optional[Reference] = None,
    ) -> RunOutput:
        """
        Creates a run.

        :param data: The RunInput json structure
        :param job: The job name or id
        :param project: The project name or id
        :param workspace: The workspace name or id

        :return: The RunOutput json structure
        """
        if isinstance(job, str):
            if project is None and self.project is not None:
                project = self.project.id
            if workspace is None and self.workspace is not None:
                workspace = self.workspace.id
        return RunApi(self._auth).create_run(data, job, project, workspace)

    def get_run(
        self,
        run: Reference,
        job: Optional[Reference] = None,
        project: Optional[Reference] = None,
        workspace: Optional[Reference] = None,
    ) -> RunOutput:
        """
        Gets a run.

        :param run: The run name or id
        :param job: The job name or id
        :param project: The project name or id
        :param workspace: The workspace name or id

        :return: The RunOutput json structure
        """
        if isinstance(job, str):
            if project is None and self.project is not None:
                project = self.project.id
            if workspace is None and self.workspace is not None:
                workspace = self.workspace.id
        return RunApi(self._auth).get_run(run, job, project, workspace)

    def list_artifacts(
        self,
        run: Reference,
        job: Optional[Reference] = None,
        project: Optional[Reference] = None,
        workspace: Optional[Reference] = None,
    ) -> PagedResponse[ArtifactOutput]:
        """
        Lists the artifacts of a run.

        :param run: The run name or id
        :param job: The job name or id
        :param project: The project name or id
        :param workspace: The workspace name or id

        :return: PagedResponse[ArtifactOutput]
        """

        if isinstance(job, str):
            if project is None and self.project is not None:
                project = self.project.id
            if workspace is None and self.workspace is not None:
                workspace = self.workspace.id
        return ArtifactApi(self._auth).list_artifacts(run, job, project, workspace)

    def update_run(self, data: RunOutput):
        """
        Updates a run.

        :param data: The RunOutput json sturcture

        :return: The RunOutput json structure
        """
        if data.id is None:
            raise ValueError("the id is required")
        return RunApi(self._auth).update_run(data)

    def delete_run(
        self,
        run: Reference,
        job: Optional[Reference] = None,
        project: Optional[Reference] = None,
        workspace: Optional[Reference] = None,
    ):
        """
        Deletes a run.

        :param run: The run name or id
        :param job: The job name or id
        :param project: The project name or id
        :param workspace: The workspace name or id

        :return: None
        """
        RunApi(self._auth).delete_run(run, job, project, workspace)

    def create_run_properties(
        self,
        run: Reference,
        properties: List[PropertyInput],
        job: Optional[Reference] = None,
        project: Optional[Reference] = None,
        workspace: Optional[Reference] = None,
    ) -> RunOutput:
        """
        Creates run properties.

        :param run: The run name or id
        :param properties: The list of PropertyInput json structure
        :param job: The job name or id
        :param project: The project name or id
        :param workspace: The workspace name or id

        :return: The RunOutput json structure
        """
        if isinstance(job, str) and isinstance(run, str):
            if project is None and self.project is not None:
                project = self.project.id
            elif isinstance(project, str):
                if workspace is None and self.workspace is not None:
                    workspace = self.workspace.id
        return RunApi(self._auth).create_run_properties(run, properties, job, project, workspace)

    def list_run_properties(
        self,
        run: Reference,
        job: Optional[Reference] = None,
        project: Optional[Reference] = None,
        workspace: Optional[Reference] = None,
        page_index=Page.index,
        page_size=Page.size,
    ) -> PagedResponse[PropertyOutput]:
        """
        Lists the run properties.

        :param run: The run name or id
        :param job: The job name or id
        :param project: The project name or id
        :param workspace: The workspace name or id
        :param page_index: The index of the page
        :param page_size: The size of the page

        :return: PagedResponse[PropertyOutput]
        """
        if isinstance(job, str) and isinstance(run, str):
            if project is None and self.project is not None:
                project = self.project.id
            elif isinstance(project, str):
                if workspace is None and self.workspace is not None:
                    workspace = self.workspace.id
        return RunApi(self._auth).list_run_properties(
            run,
            job,
            project,
            workspace,
            page_index,
            page_size,
        )

    def create_artifact(
        self,
        data: ArtifactInput,
        run: Reference,
        job: Optional[Reference] = None,
        project: Optional[Reference] = None,
        workspace: Optional[Reference] = None,
    ):
        """
        Creates an artifact.

        :param data: The ArtifactInput json structure
        :param run: The run name or id
        :param job: The job name or id
        :param project: The project name or id
        :param workspace: The workspace name or id

        :return: The ArtifactOutput json structure
        """
        return ArtifactApi(self._auth).create_artifact(data, run, job, project, workspace)

    def update_artifact(self, artifact: ArtifactOutput):
        """
        Updates an artifact.

        :param artifact: The ArtifactOutput json structure

        :return: The json structure
        """
        return ArtifactApi(self._auth).update_artifact(artifact)

    def get_dataset(
        self, dataset: Reference, project: Optional[Reference] = None, workspace: Optional[Reference] = None
    ) -> DatasetOutput:
        """
        Gets a dataset.

        :param dataset: The dataset name or id
        :param project: The project name or id
        :param workspace: The workspace name or id

        :return: The DatasetOutput json structure
        """
        if isinstance(dataset, str):
            if project is None and self.project is not None:
                project = self.project.id
            elif isinstance(project, str):
                if workspace is None and self.workspace is not None:
                    workspace = self.workspace.id
        return DatasetApi(self._auth).get_dataset(dataset, project, workspace)

    def list_datasets(
        self,
        project: Optional[Reference] = None,
        workspace: Optional[Reference] = None,
        search: str = None,
        page_index=Page.index,
        page_size=Page.size,
    ) -> PagedResponse[DatasetOutput]:
        """
        Lists the datasets in a project.

        :param project: The project name or id
        :param workspace: The workspace name or id
        :param search: A name to search for
        :param page_index: The index of the page
        :param page_size: The size of the page

        :return: PagedResponse[DatasetOutput]
        """
        if project is None and self.project is not None:
            project = self.project.id
        elif isinstance(project, str):
            if workspace is None and self.workspace is not None:
                workspace = self.workspace.id
        if project is None:
            raise MissingReferenceError("project")
        return DatasetApi(self._auth).list_datasets(project, workspace, search, page_index, page_size)

    def create_dataset(
        self, data: DatasetInput, project: Optional[Reference] = None, workspace: Optional[Reference] = None
    ) -> DatasetOutput:
        """
        Creates a dataset.

        :param data: The DatasetOutput json structure
        :param project: The project name or id
        :param workspace: The workspace name or id

        :return: The DatasetOutput json structure
        """
        if project is None and self.project is not None:
            project = self.project.id
        elif isinstance(project, str):
            if workspace is None and self.workspace is not None:
                workspace = self.workspace.id
        if project is None:
            raise MissingReferenceError("project")
        return DatasetApi(self._auth).create_dataset(data, project, workspace)

    def update_dataset(
        self,
        data: DatasetOutput,
        dataset: Reference,
        project: Optional[Reference] = None,
        workspace: Optional[Reference] = None,
    ) -> DatasetOutput:
        """
        Updates a dataset.

        :param data: The DatasetOutput json structure
        :param dataset: Teh dataset name or id
        :param project: The project name or id
        :param workspace: The workspace name or id

        :return: The DatasetOutput json structure
        """
        if isinstance(dataset, str):
            if project is None and self.project is not None:
                project = self.project.id
            elif isinstance(project, str):
                if workspace is None and self.workspace is not None:
                    workspace = self.workspace.id
        return DatasetApi(self._auth).update_dataset(data, dataset, project, workspace)

    def delete_dataset(
        self, dataset: Reference, project: Optional[Reference] = None, workspace: Optional[Reference] = None
    ) -> None:
        """
        Deletes a dataset.

        :param dataset: The dataset name or id
        :param project: The project name or id
        :param workspace: The workspace name or id

        :return: None
        """
        if isinstance(dataset, str):
            if project is None and self.project is not None:
                project = self.project.id
            elif isinstance(project, str):
                if workspace is None and self.workspace is not None:
                    workspace = self.workspace.id
        DatasetApi(self._auth).delete_dataset(dataset, project, workspace)

    def get_model(
        self,
        model: Reference,
        project: Optional[Reference] = None,
        workspace: Optional[Reference] = None,
    ):
        """
        Gets a model.

        :param model: The model name or id
        :param project: The project name or id
        :param workspace: The workspace name or id

        :return: The ModelOutput json structure
        """
        if isinstance(model, str):
            if project is None and self.project is not None:
                project = self.project.id
            elif isinstance(project, str):
                if workspace is None and self.workspace is not None:
                    workspace = self.workspace.id
        return ModelApi(self._auth).get_model(model, project, workspace)

    def list_models(
        self,
        project: Optional[Reference] = None,
        workspace: Optional[Reference] = None,
        search: Optional[str] = None,
        page_index=Page.index,
        page_size=Page.size,
    ) -> PagedResponse[ModelOutput]:
        """
        Lists the models in a project.

        :param project: The project name or id
        :param workspace: The workspace name or id
        :param search: The name to search
        :param page_index: The index of the page
        :param page_size: The size of the page

        :return: PagedResponse[ModelOutput]
        """
        if project is None and self.project is not None:
            project = self.project.id
        elif isinstance(project, str):
            if workspace is None and self.workspace is not None:
                workspace = self.workspace.id
        if project is None:
            raise MissingReferenceError("project")
        return ModelApi(self._auth).list_models(project, workspace, search, page_index, page_size)

    def create_model(
        self, data: ModelInput, project: Optional[Reference] = None, workspace: Optional[Reference] = None
    ):
        """
        Creates a model.

        :param data: The ModelInput json structure
        :param project: The project name or id
        :param workspace: The workspace name or id

        :return: The ModelOutput json structure
        """
        if project is None and self.project is not None:
            project = self.project.id
        if workspace is None and self.workspace is not None:
            workspace = self.workspace.id
        if project is None:
            raise MissingReferenceError("project")
        return ModelApi(self._auth).create_model(data, project, workspace)

    def update_model(
        self,
        data: ModelInput,
        model: Reference,
        project: Optional[Reference] = None,
        workspace: Optional[Reference] = None,
    ):
        """
        Updates a model.

        :param data: The ModelInput json structure
        :param model: The model name or id
        :param project: The project name or id
        :param workspace: The workspace name or id

        :return: The ModelOutput json structure

        """
        if isinstance(model, str):
            if project is None and self.project is not None:
                project = self.project.id
            elif isinstance(project, str):
                if workspace is None and self.workspace is not None:
                    workspace = self.workspace.id
        return ModelApi(self._auth).update_model(data, model, project, workspace)

    def delete_model(
        self,
        model: Reference,
        project: Optional[Reference] = None,
        workspace: Optional[Reference] = None,
    ):
        """
        Deletes a model.

        :param model: The model name or id
        :param project: The project name or id
        :param workspace: The workspace name or id

        :return: None

        """
        ModelApi(self._auth).delete_model(model, project, workspace)

    def list_dataset_versions(
        self,
        dataset: Reference,
        search: Optional[str] = None,
        page_index=Page.index,
        page_size=Page.size,
        project: Optional[Reference] = None,
        workspace: Optional[Reference] = None,
    ) -> PagedResponse[DatasetVersionOutput]:
        """
        Lists dataset versions.

        :param dataset: The dataset name or id
        :param search: A name to search
        :param page_index: The index of the page
        :param page_size: The size of the page
        :param project: The project name or id
        :param workspace: the workspace name or id

        :return: PagedResponse[DatasetVersionOutput]
        """
        if isinstance(dataset, str):
            if project is None and self.project is not None:
                project = self.project.id
            elif isinstance(project, str):
                if workspace is None and self.workspace is not None:
                    workspace = self.workspace.id
        return DatasetVersionApi(self._auth).list_dataset_versions(
            dataset, workspace, project, search, page_index, page_size
        )

    def create_dataset_version(
        self,
        data: DatasetVersionInput,
        dataset: Reference,
        project: Optional[Reference] = None,
        workspace: Optional[Reference] = None,
    ) -> DatasetVersionOutput:
        """
        Creates a dataset version.

        :param data: The DatasetVersionOutput json structure
        :param dataset: The parent dataset name or id
        :param project: The project name or id
        :param workspace: the workspace name or id

        :return: The DatasetVersionOutput json structure
        """
        if isinstance(dataset, str):
            if project is None and self.project is not None:
                project = self.project.id
            elif isinstance(project, str):
                if workspace is None and self.workspace is not None:
                    workspace = self.workspace.id
        return DatasetVersionApi(self._auth).create_dataset_version(data, dataset, project, workspace)

    def update_dataset_version(
        self,
        data: DatasetVersionOutput,
        version: Reference,
        dataset: Optional[Reference] = None,
        project: Optional[Reference] = None,
        workspace: Optional[Reference] = None,
    ) -> DatasetVersionOutput:
        """
        Updates a dataset version.

        :param data: The DatasetVersionOutput json structure
        :param version: The version name or id
        :param dataset: The parent dataset name or id
        :param project: The project name or id
        :param workspace: the workspace name or id

        :return: The DatasetVersionOutput json structure
        """
        if isinstance(version, str) and isinstance(dataset, str):
            if project is None and self.project is not None:
                project = self.project.id
            elif isinstance(project, str):
                if workspace is None and self.workspace is not None:
                    workspace = self.workspace.id
        return DatasetVersionApi(self._auth).update_dataset_version(data, version, dataset, project, workspace)

    def get_dataset_version(
        self,
        version: Reference,
        dataset: Optional[Reference] = None,
        project: Optional[Reference] = None,
        workspace: Optional[Reference] = None,
    ) -> DatasetVersionOutput:
        """
        Gets a dataset version.

        :param version: The version name or id
        :param dataset: The parent dataset name or id
        :param project: The project name or id
        :param workspace: the workspace name or id

        :return: DatasetVersionOutput json structure
        """
        if isinstance(version, str) and isinstance(dataset, str):
            if project is None and self.project is not None:
                project = self.project.id
            elif isinstance(project, str):
                if workspace is None and self.workspace is not None:
                    workspace = self.workspace.id
        return DatasetVersionApi(self._auth).get_dataset_version(version, dataset, project, workspace)

    def delete_dataset_version(
        self,
        version: Reference,
        dataset: Optional[Reference] = None,
        project: Optional[Reference] = None,
        workspace: Optional[Reference] = None,
    ):
        """
        Deletes a dataset version.

        :param version: The version name or id
        :param dataset: The parent dataset name or id
        :param project: The project name or id
        :param workspace: the workspace name or id

        :return: None
        """
        if isinstance(version, str) and isinstance(dataset, str):
            if project is None and self.project is not None:
                project = self.project.id
            elif isinstance(project, str):
                if workspace is None and self.workspace is not None:
                    workspace = self.workspace.id
        DatasetVersionApi(self._auth).delete_dataset_version(version, dataset, project, workspace)

    def list_dataset_version_files_metadata(
        self,
        version: Reference,
        dataset: Optional[Reference] = None,
        project: Optional[Reference] = None,
        workspace: Optional[Reference] = None,
    ) -> List[FileMetadata]:
        """
        Lists the dataset version properties.

        :param version: The dataset version name or ID
        :param dataset: The parent dataset name or ID
        :param project: The project name or ID
        :param workspace: The workspace name or ID

        :return: DatasetVersionOutput json structure
        """
        return DatasetVersionApi(self._auth).list_files_metadata(version, dataset, project, workspace)

    def create_dataset_version_properties(
        self,
        version: Reference,
        properties: Dict[str, Any],
        dataset: Optional[Reference] = None,
        project: Optional[Reference] = None,
        workspace: Optional[Reference] = None,
    ):
        """
        Creates dataset version properties.

        :param version: The dataset version name or ID
        :param properties: The properties dictionary
        :param dataset: The parent dataset name or ID
        :param project: The project name or ID
        :param workspace: The workspace name or ID

        :return: DatasetVersionOutput json structure
        """
        if isinstance(version, str) and isinstance(dataset, str):
            if project is None and self.project is not None:
                project = self.project.id
            elif isinstance(project, str):
                if workspace is None and self.workspace is not None:
                    workspace = self.workspace.id
        result = DatasetVersionApi(self._auth).create_dataset_version_properties(
            version, properties, dataset, project, workspace
        )
        self._logger.info(
            f"Properties with names: {list(properties.keys())} successfully added to DatasetVersion {result.name}."
        )
        return result

    def update_dataset_version_properties(
        self,
        version: Reference,
        property_id: int,
        properties: Dict[str, Any],
        dataset: Optional[Reference] = None,
        project: Optional[Reference] = None,
        workspace: Optional[Reference] = None,
    ):
        """
        Updates dataset version properties.

        :param version: The dataset version name or id
        :param property_id: The property key or id to update
        :param properties: The properties dictionary
        :param dataset: The parent dataset name or id
        :param project: The project name or id
        :param workspace: The workspace name or id

        :return: None
        """
        if isinstance(version, str) and isinstance(dataset, str):
            if project is None and self.project is not None:
                project = self.project.id
            elif isinstance(project, str):
                if workspace is None and self.workspace is not None:
                    workspace = self.workspace.id
        result = DatasetVersionApi(self._auth).update_dataset_version_properties(
            version, property_id, properties, dataset, project, workspace
        )
        dataset_version = DatasetVersionApi(self._auth).get_dataset_version(version, dataset, project, workspace)
        self._logger.info(
            f"Properties with names: {list(properties.keys())} successfully updated in DatasetVersion {dataset_version.name}."
        )
        return result

    def list_dataset_version_properties(
        self,
        version: Reference,
        dataset: Optional[Reference] = None,
        project: Optional[Reference] = None,
        workspace: Optional[Reference] = None,
        page_index=Page.index,
        page_size=Page.size,
    ) -> PagedResponse[PropertyOutput]:
        """
        Lists dataset version properties.

        :param version: The dataset version name or id
        :param page_index: The index of the page
        :param page_size: The size of the page
        :param dataset: Teh parent dataset name or id
        :param project: The project name or id
        :param workspace: The workspace name or id

        :return: PagedResponse[PropertyOutput]
        """
        if isinstance(version, str) and isinstance(dataset, str):
            if project is None and self.project is not None:
                project = self.project.id
            elif isinstance(project, str):
                if workspace is None and self.workspace is not None:
                    workspace = self.workspace.id
        return DatasetVersionApi(self._auth).list_dataset_version_properties(
            version,
            dataset,
            project,
            workspace,
            page_index,
            page_size,
        )

    def get_model_version(
        self,
        version: Reference,
        model: Optional[Reference] = None,
        project: Optional[Reference] = None,
        workspace: Optional[Reference] = None,
    ) -> ModelVersionOutput:
        """
        Gets a model version.

        :param version: The model version name or id
        :param model: The model name or id
        :param project: The project name or id
        :param workspace: The workspace name or id

        :return: ModelVersionOutput
        """
        if isinstance(version, str) and isinstance(model, str):
            if project is None and self.project is not None:
                project = self.project.id
            elif isinstance(project, str):
                if workspace is None and self.workspace is not None:
                    workspace = self.workspace.id
        return ModelVersionApi(self._auth).get_model_version(version, model, project, workspace)

    def list_model_versions(
        self,
        model: Reference,
        project: Optional[Reference] = None,
        workspace: Optional[Reference] = None,
        search: Optional[str] = None,
        page_index=Page.index,
        page_size=Page.size,
    ) -> PagedResponse[ModelVersionOutput]:
        """
        Lists model versions.

        :param model: The model name or id
        :param project: The project name or id
        :param workspace: The workspace name or id
        :param search: The word to search
        :param page_index: The index of the page
        :param page_size: The size of the page

        :return: PagedResponse[ModelVersionOutput]
        """
        if isinstance(model, str):
            if project is None and self.project is not None:
                project = self.project.id
            elif isinstance(project, str):
                if workspace is None and self.workspace is not None:
                    workspace = self.workspace.id
        return ModelVersionApi(self._auth).list_model_versions(model, project, workspace, search, page_index, page_size)

    def list_model_version_metrics(
        self,
        version: Reference,
        model: Optional[Reference] = None,
        project: Optional[Reference] = None,
        workspace: Optional[Reference] = None,
        search: Optional[str] = None,
        page_index=Page.index,
        page_size=Page.size,
    ) -> PagedResponse[MetricOutput]:
        """
        Lists model versions.

        :param model: The model name or id
        :param version: The model version name or id
        :param project: The project name or id
        :param workspace: The workspace name or id
        :param search: The word to search
        :param page_index: The index of the page
        :param page_size: The size of the page

        :return:PagedResponse[MetricOutput]
        """
        if isinstance(version, str) and isinstance(model, str):
            if project is None and self.project is not None:
                project = self.project.id
            elif isinstance(project, str) and self.workspace is not None:
                if workspace is None:
                    workspace = self.workspace.id
        return ModelVersionApi(self._auth).list_model_version_metrics(
            version,
            model,
            project,
            workspace,
            search,
            page_index,
            page_size,
        )

    def list_model_version_properties(
        self,
        version: Reference,
        model: Optional[Reference] = None,
        project: Optional[Reference] = None,
        workspace: Optional[Reference] = None,
        search: Optional[str] = None,
        page_index=Page.index,
        page_size=Page.size,
    ) -> PagedResponse[PropertyOutput]:
        """
        Lists model version properties.

        :param model: The model name or id
        :param version: The model version name or id
        :param project: The project name or id
        :param workspace: The workspace name or id
        :param search: The word to search
        :param page_index: The index of the page
        :param page_size: The size of the page

        :return: PagedResponse[PropertyOutput]
        """
        if isinstance(version, str):
            if project is None and self.project is not None:
                project = self.project.id
            if workspace is None and self.workspace is not None:
                workspace = self.workspace.id
        return ModelVersionApi(self._auth).list_model_version_properties(
            version, model, project, workspace, search, page_index, page_size
        )

    def create_model_version_properties(
        self,
        version: Reference,
        properties: List[PropertyInput],
        model: Optional[Reference] = None,
        project: Optional[Reference] = None,
        workspace: Optional[Reference] = None,
    ):
        """
        Creates model version properties.

        :param model: The model name or id
        :param version: The model version name or id
        :param properties: The properties object
        :param project: The project name or id
        :param workspace: The workspace name or id

        :return: None
        """
        if project is None and self.project is not None:
            project = self.project.id
        if workspace is None and self.workspace is not None:
            workspace = self.workspace.id
        return ModelVersionApi(self._auth).create_model_version_properties(
            version, properties, model, project, workspace
        )

    def update_model_version_properties(
        self,
        version: Reference,
        property_id: int,
        properties: List[PropertyInput],
        model: Optional[Reference] = None,
        project: Optional[Reference] = None,
        workspace: Optional[Reference] = None,
    ):
        """
        Creates model version properties.

        :param model: The model name or id
        :param version: The model version name or id
        :param property_id: The property id
        :param properties: The properties object
        :param project: The project name or id
        :param workspace: The workspace name or id

        :return: None

        """
        if project is None and self.project is not None:
            project = self.project.id
        if workspace is None and self.workspace is not None:
            workspace = self.workspace.id
        return ModelVersionApi(self._auth).update_model_version_properties(
            version,
            property_id,
            properties,
            model,
            project,
            workspace,
        )

    def create_model_version_metrics(
        self,
        version: Reference,
        metrics: List[MetricInput],
        model: Optional[Reference] = None,
        project: Optional[Reference] = None,
        workspace: Optional[Reference] = None,
    ) -> None:
        """
        Creates model version metrics.

        :param model: The model name or id
        :param version: The model version name or id
        :param metrics: The properties object
        :param project: The project name or id
        :param workspace: The workspace name or id

        :return: None
        """
        if project is None and self.project is not None:
            project = self.project.id
        if workspace is None and self.workspace is not None:
            workspace = self.workspace.id
        return ModelVersionApi(self._auth).create_model_version_metrics(version, metrics, model, project, workspace)

    def update_model_version_metrics(
        self,
        model: Reference,
        version: Reference,
        metric_id: int,
        metrics: List[MetricInput],
        project: Optional[Reference] = None,
        workspace: Optional[Reference] = None,
    ):
        """
        Creates model version metrics.

        :param model: The model name or id
        :param version: The model version name or id
        :param metric_id: The metrics id
        :param metrics: The metrics object
        :param project: The project name or id
        :param workspace: The workspace name or id

        :return: None
        """
        if project is None and self.project is not None:
            project = self.project.id
        if workspace is None and self.workspace is not None:
            workspace = self.workspace.id
        return ModelVersionApi(self._auth).update_model_version_metrics(
            version,
            metric_id,
            metrics,
            model,
            project,
            workspace,
        )

    def create_model_version(
        self,
        data: ModelVersionInput,
        model: Reference,
        project: Optional[Reference] = None,
        workspace: Optional[Reference] = None,
    ) -> ModelVersionOutput:
        """
        Creates a model version.

        :param data: The ModelVersionInput json structure
        :param model: The model name or id
        :param project: The project name or id
        :param workspace: The workspace name or id

        :return: The ModelVersionOutput json structure
        """
        if project is None and self.project is not None:
            project = self.project.id
        if workspace is None and self.workspace is not None:
            workspace = self.workspace.id
        return ModelVersionApi(self._auth).create_model_version(data, model, project, workspace)

    def update_model_version(
        self,
        data: ModelVersionInput,
        version: Reference,
        model: Optional[Reference] = None,
        project: Optional[Reference] = None,
        workspace: Optional[Reference] = None,
    ) -> ModelVersionOutput:
        """
        Updates a model version.

        :param data: The ModelVersionInput json structure
        :param version: The model version name or id
        :param model: The model name or id
        :param project: The project name or id
        :param workspace: The workspace name or id

        :return: The ModelVersionOutput json structure

        """
        if project is None and self.project is not None:
            project = self.project.id
        if workspace is None and self.workspace is not None:
            workspace = self.workspace.id
        return ModelVersionApi(self._auth).update_model_version(data, version, model, project, workspace)

    def delete_model_version(
        self,
        version: Reference,
        model: Optional[Reference] = None,
        project: Optional[Reference] = None,
        workspace: Optional[Reference] = None,
    ):
        """
        Deletes a dataset version.

        :param version: The model version name or id
        :param model: The model name or id
        :param project: The project name or id
        :param workspace: The workspace name or id

        :return: None
        """
        if isinstance(version, str) and isinstance(model, str):
            if project is None and self.project is not None:
                project = self.project.id
            elif isinstance(project, str):
                if workspace is None and self.workspace is not None:
                    workspace = self.workspace.id
        ModelVersionApi(self._auth).delete_model_version(version, model, project, workspace)

    def delete_model_version_metrics(
        self,
        version: Reference,
        metric_id: int,
        model: Reference,
        project: Optional[Reference] = None,
        workspace: Optional[Reference] = None,
    ):
        """
        Deletes a model version metric.

        :param version: The model version name or id
        :param metric_id: The metrics id
        :param model: The model name or id
        :param project: The project name or id
        :param workspace: The workspace name or id

        :return: None
        """
        if isinstance(version, str) and isinstance(model, str):
            if project is None and self.project is not None:
                project = self.project.id
            elif isinstance(project, str):
                if workspace is None and self.workspace is not None:
                    workspace = self.workspace.id
        return ModelVersionApi(self._auth).delete_model_version_metrics(version, metric_id, model, project, workspace)

    def delete_model_version_properties(
        self,
        version: Reference,
        property_id: int,
        model: Reference,
        project: Optional[Reference] = None,
        workspace: Optional[Reference] = None,
    ):
        """
        Deletes a model version property.

        :param version: The model version name or id
        :param property_id: The property id
        :param model: The model name or id
        :param project: The project name or id
        :param workspace: The workspace name or id
        :return: None
        """
        if isinstance(version, str) and isinstance(model, str):
            if project is None and self.project is not None:
                project = self.project.id
            elif isinstance(project, str):
                if workspace is None and self.workspace is not None:
                    workspace = self.workspace.id
        return ModelVersionApi(self._auth).delete_model_version_properties(
            version, property_id, model, project, workspace
        )

    def get_attachment(
        self,
        _type: str,
        file_id: int,
        version: Optional[Reference] = None,
        artifact: Optional[Reference] = None,
        workspace: Optional[Reference] = None,
        project: Optional[Reference] = None,
    ) -> BinaryIO:
        """
        Downloads the specified attachment.

        :param _type: Artifact type of the attachment, it can be "datasetversion", "codeversion" or "modelversion"
        :param file_id: The file id
        :param version: The artifact version name or id
        :param artifact: The artifact name or id
        :param workspace: The workspace name or id
        :param project: The project name or id

        :return: BinaryIO
        """
        if project is None and self.project is not None:
            project = self.project.id
        elif isinstance(project, str):
            if workspace is None and self.workspace is not None:
                workspace = self.workspace.id
        return AttachmentApi(self._auth).get_attachment(_type, file_id, version, artifact, workspace, project)

    def create_attachments(
        self,
        _type: str,
        files: Optional[Sequence[Tuple[str, Tuple[Any, BinaryIO]]]],
        version: Optional[Reference] = None,
        artifact: Optional[Reference] = None,
        workspace: Optional[Reference] = None,
        project: Optional[Reference] = None,
    ):
        """
        Creates an attachment

        :param _type: Artifact type of the attachment, it can be "datasetversion", "codeversion" or "modelversion"
        :param files: The paths to the files to attach
        :param version: The artifact version name or id
        :param artifact: The artifact name or id
        :param workspace: The workspace name or id
        :param project: The project name or id

        :return: The json structure
        """
        if project is None and self.project is not None:
            project = self.project.id
        elif isinstance(project, str):
            if workspace is None and self.workspace is not None:
                workspace = self.workspace.id
        return AttachmentApi(self._auth).post_attachment(_type, version, files, artifact, workspace, project)

    def update_attachments(
        self,
        _type: str,
        files: Sequence[Tuple[str, Tuple[Any, BinaryIO]]],
        version: Optional[Reference] = None,
        artifact: Optional[Reference] = None,
        workspace: Optional[Reference] = None,
        project: Optional[Reference] = None,
    ):
        """
        Updates an attachment.

        :param _type: Artifact type of the attachment, it can be "datasetversion", "codeversion" or "modelversion"
        :param files: The paths to the files to attach
        :param version: The artifact version name or id
        :param artifact: The artifact name or id
        :param workspace: The workspace name or id
        :param project: The project name or id

        :return: None
        """
        if project is None and self.project is not None:
            project = self.project.id
        elif isinstance(project, str):
            if workspace is None and self.workspace is not None:
                workspace = self.workspace.id
        return AttachmentApi(self._auth).update_attachments(_type, files, version, artifact, workspace, project)

    def delete_attachment(
        self,
        _type: str,
        version: Reference,
        file_id: int,
        artifact: Optional[Reference] = None,
        workspace: Optional[Reference] = None,
        project: Optional[Reference] = None,
    ):
        """
        Deletes an attachment.

        :param _type: Artifact type of the attachment, it can be "datasetversion", "codeversion" or "modelversion"
        :param version: The artifact version name or id
        :param file_id: The file id
        :param artifact: The artifact name or id
        :param workspace: The workspace name or id
        :param project: The project name or id


        :return: None
        """
        if project is None and self.project is not None:
            project = self.project.id
        elif isinstance(project, str):
            if workspace is None and self.workspace is not None:
                workspace = self.workspace.id
        return AttachmentApi(self._auth).delete_attachment(_type, version, file_id, artifact, workspace, project)

    def list_attachments(
        self,
        _type: str,
        version: Optional[Reference] = None,
        artifact: Optional[Reference] = None,
        workspace: Optional[Reference] = None,
        project: Optional[Reference] = None,
    ) -> PagedResponse[AttachmentOutput]:
        """
        Lists the attachments of an artifact.

        :param _type: Artifact type of the attachment, it can be "datasetversion", "codeversion" or "modelversion"
        :param version: The artifact version name or id
        :param artifact: The artifact name or id
        :param workspace: The workspace name or id
        :param project: The project name or id

        :return: PagedResponse[AttachmentOutput]
        """
        if project is None and self.project is not None:
            project = self.project.id
        elif isinstance(project, str):
            if workspace is None and self.workspace is not None:
                workspace = self.workspace.id
        return AttachmentApi(self._auth).list_attachments(_type, version, artifact, workspace, project)

    def get_connection(self, connection: Reference, workspace: Optional[Reference] = None) -> ConnectionOutput:
        """
        Gets the specified connection in a workspace.

        :param connection: The connection name or id
        :param workspace: The workspace name or id

        :return: The ConnectionOutput json structure
        """

        if workspace is None and self.workspace is not None:
            workspace = self.workspace.id
        if workspace is None:
            raise MissingReferenceError("workspace")
        return ConnectionApi(self._auth).get_connection(connection, workspace)

    def list_connections(
        self,
        workspace: Optional[Reference] = None,
        connection_type: Optional[str] = None,
        search: Optional[str] = None,
        page_index: int = Page.index,
        page_size: int = Page.size,
    ) -> PagedResponse[ConnectionOutput]:
        """
        Lists the connections in a workspace.

        :param workspace: The workspace name or id
        :param search: A name to search for
        :param connection_type: The connection type to search for
        :param page_index: The page index
        :param page_size: The page size

        :return: PagedResponse[ConnectionOutput]
        """
        if workspace is None and self.workspace is not None:
            workspace = self.workspace.id
        if workspace is None:
            raise MissingReferenceError("workspace")
        return ConnectionApi(self._auth).list_connections(workspace, connection_type, search, page_index, page_size)

    def create_connection(self, data: ConnectionInput, workspace: Optional[Reference] = None) -> ConnectionOutput:
        """
        Creates a connection in a workspace.

        :param data: The ConnectionInput json structure
        :param workspace: The workspace name or id

        :return: The ConnectionOutput json structure
        """
        if workspace is None and self.workspace is not None:
            workspace = self.workspace.id
        if workspace is None:
            raise MissingReferenceError("workspace")
        return ConnectionApi(self._auth).create_connection(data, workspace)

    def delete_connection(self, connection: Reference, workspace: Optional[Reference] = None):
        """
        Deletes a connection from a workspace.

        :param connection: The connection name or id
        :param workspace: The workspace name or id

        :return: None
        """
        if workspace is None and self.workspace is not None:
            workspace = self.workspace.id
        if workspace is None:
            raise MissingReferenceError("workspace")
        ConnectionApi(self._auth).delete_connection(connection, workspace)

    def update_connection(
        self, data: ConnectionInput, connection: Reference, workspace: Optional[Reference] = None
    ) -> ConnectionOutput:
        """
        Updates a connection in a workspace.

        :param data: The new ConnectionInput json structure
        :param connection: The connection name or id
        :param workspace: The workspace name or id

        :return: The ConnectionOutput json structure
        """
        if workspace is None and self.workspace is not None:
            workspace = self.workspace.id
        if workspace is None:
            raise MissingReferenceError("workspace")
        return ConnectionApi(self._auth).update_connection(data, connection, workspace)

    def create_code(
        self,
        code_data: CodeInput,
        project: Optional[Reference] = None,
        workspace: Optional[Reference] = None,
    ) -> CodeOutput:
        """
        Creates a code in a project.

        :param code_data: The CodeInput json structure
        :param project: The project name or id
        :param workspace: The workspace name or id

        :return: The CodeOutput json structure
        """
        if project is None and self.project is not None:
            project = self.project.id
        elif isinstance(project, str):
            if workspace is None and self.workspace is not None:
                workspace = self.workspace.id
        if project is None:
            raise MissingReferenceError("project")
        result = CodeApi(self._auth).create_code(code_data, project, workspace)
        self._logger.info(f"Code with id: {result.id} successfully created.")
        return result

    def get_code(
        self,
        code: Reference,
        project: Optional[Reference] = None,
        workspace: Optional[Reference] = None,
    ) -> CodeOutput:
        """
        Gets a code in a project.

        :param code: The code name or id
        :param project: The project name or id
        :param workspace: The workspace name or id

        :return: The CodeOutput json structure
        """
        if project is None and self.project is not None:
            project = self.project.id
        elif isinstance(project, str):
            if workspace is None and self.workspace is not None:
                workspace = self.workspace.id
        result = CodeApi(self._auth).get_code(code, project, workspace)
        self._logger.info(f"Code with id: {result.id} successfully retrieved.")
        return result

    def list_codes(
        self,
        project: Reference,
        workspace: Optional[Reference] = None,
        search: Optional[str] = None,
        page_index=Page.index,
        page_size=Page.size,
    ) -> PagedResponse[CodeOutput]:
        """
        Lists the codes in a project.

        :param project: The project name or id
        :param workspace: The workspace name or id
        :param search: A name to search for
        :param page_index: The page index
        :param page_size: The page size

        :return: PagedResponse[CodeOutput]
        """
        if project is None and self.project is not None:
            project = self.project.id
        elif isinstance(project, str):
            if workspace is None and self.workspace is not None:
                workspace = self.workspace.id
        return CodeApi(self._auth).list_codes(project, workspace, search, page_index, page_size)

    def update_code(
        self,
        code_data: CodeOutput,
        project: Optional[Reference] = None,
        workspace: Optional[Reference] = None,
    ):
        """
        Updates a code in a project.

        :param code_data: The CodeOutput json structure
        :param project: The project name or id
        :param workspace: The workspace name or id

        :return: The CodeOutput json structure
        """
        if project is None and self.project is not None:
            project = self.project.id
        elif isinstance(project, str):
            if workspace is None and self.workspace is not None:
                workspace = self.workspace.id
        result = CodeApi(self._auth).update_code(code_data, project, workspace)
        self._logger.info(f"Code with id: {result.id} successfully updated.")
        return result

    def delete_code(
        self,
        code: Reference,
        project: Optional[Reference] = None,
        workspace: Optional[Reference] = None,
    ) -> None:
        """
        Deletes a code in a project.

        :param code: The code name or id
        :param project: The project name or id
        :param workspace: The workspace name or id

        :return: None
        """
        if project is None and self.project is not None:
            project = self.project.id
        elif isinstance(project, str):
            if workspace is None and self.workspace is not None:
                workspace = self.workspace.id
        code_id = CodeApi(self._auth).get_code(code, project, workspace).id
        CodeApi(self._auth).delete_code(code, project, workspace)
        self._logger.info(f"Code with id: {code_id} successfully deleted.")

    def create_code_version(
        self,
        code_data: CodeVersionInput,
        project: Optional[Reference] = None,
        workspace: Optional[Reference] = None,
    ) -> CodeVersionOutput:
        """
        Creates a code version.

        :param code_data: The CodeVersionInput json structure
        :param project: The project name or id
        :param workspace: The workspace name or id

        :return: The CodeVersionOutput json structure
        """
        if project is None and self.project is not None:
            project = self.project.id
        if workspace is None and self.workspace is not None:
            workspace = self.workspace.id
        return CodeVersionApi(self._auth).create_code_version(code_data, project, workspace)

    def get_code_version(
        self,
        version: Reference,
        project: Optional[Reference] = None,
        workspace: Optional[Reference] = None,
    ) -> CodeVersionOutput:
        """
        Gets a code version.

        :param version: The code version name or id
        :param project: The project name or id
        :param workspace: The workspace name or id

        :return: The CodeVersionOutput json structure
        """

        if project is None and self.project is not None:
            project = self.project.id
        if workspace is None and self.workspace is not None:
            workspace = self.workspace.id
        return CodeVersionApi(self._auth).get_code_version(version, project, workspace)

    def list_code_versions(
        self,
        project: Optional[Reference] = None,
        workspace: Optional[Reference] = None,
        search: Optional[str] = None,
        page_index=Page.index,
        page_size=Page.size,
    ) -> List[CodeVersionOutput]:
        """
        Lists the code versions in a project.

        :param project: The project name or id
        :param workspace: The workspace name or id
        :param search: A name to search for
        :param page_index: The page index
        :param page_size: The page size

        :return: List[CodeVersionOutput]
        """

        if project is None and self.project is not None:
            project = self.project.id
        elif isinstance(project, str):
            if workspace is None and self.workspace is not None:
                workspace = self.workspace.id
        if project is None:
            raise MissingReferenceError("project")
        return CodeVersionApi(self._auth).list_code_versions(project, workspace, search, page_index, page_size)

    def update_code_version(
        self,
        code_version_data: CodeVersionInput,
        version: Reference,
        project: Optional[Reference] = None,
        workspace: Optional[Reference] = None,
    ) -> CodeVersionOutput:
        """
        Updates a code version.

        :param code_version_data: The CodeVersionInput json structure
        :param version: The code version name or id
        :param project: The project name or id
        :param workspace: The workspace name or id

        :return: The CodeVersionOutput json structure
        """

        if project is None and self.project is not None:
            project = self.project.id
        elif isinstance(project, str):
            if workspace is None and self.workspace is not None:
                workspace = self.workspace.id
        if project is None:
            raise MissingReferenceError("project")
        return CodeVersionApi(self._auth).update_code_version(code_version_data, version, project, workspace)

    def delete_code_version(
        self,
        version: Reference,
        project: Optional[Reference] = None,
        workspace: Optional[Reference] = None,
    ) -> None:
        """
        Deletes a code version.

        :param version: The code version name or id
        :param project: The project name or id
        :param workspace: The workspace name or id

        :return: None
        """
        if isinstance(version, str):
            if project is None and self.project is not None:
                project = self.project.id
            elif isinstance(project, str):
                if workspace is None and self.workspace is not None:
                    workspace = self.workspace.id
        code_version = CodeVersionApi(self._auth).get_code_version(version, project, workspace)
        version_id = code_version.id
        code_name = code_version.code.name
        CodeVersionApi(self._auth).delete_code_version(version, project, workspace)
        self._logger.info(f"CodeVersion with id: {version_id} successfully deleted from Code {code_name}.")

    def create_stage(
        self, data: StageInput, project: Optional[Reference] = None, workspace: Optional[Reference] = None
    ) -> StageOutput:
        if project is None and self.project is not None:
            project = self.project.id
        if workspace is None and self.workspace is not None:
            workspace = self.workspace.id
        if project is None:
            raise MissingReferenceError("project")
        return StageApi(self._auth).create_stage(data, project, workspace)

    def update_stage(
        self,
        stage: Reference,
        data: StageInput,
        project: Optional[Reference] = None,
        workspace: Optional[Reference] = None,
    ) -> StageOutput:
        if isinstance(stage, str):
            if project is None and self.project is not None:
                project = self.project.id
            elif isinstance(project, str):
                if workspace is None and self.workspace is not None:
                    workspace = self.workspace.id
        return StageApi(self._auth).update_stage(data, stage, project, workspace)

    def get_stage(
        self, stage: Reference, project: Optional[Reference] = None, workspace: Optional[Reference] = None
    ) -> StageOutput:
        if isinstance(stage, str):
            if project is None and self.project is not None:
                project = self.project.id
            elif isinstance(project, str):
                if workspace is None and self.workspace is not None:
                    workspace = self.workspace.id
        return StageApi(self._auth).get_stage(stage, project, workspace)

    def list_stages(
        self,
        workspace: Optional[Reference] = None,
        project: Optional[Reference] = None,
        search: Optional[str] = "",
        page_index: int = Page.index,
        page_size: int = Page.size,
    ) -> List[StageOutput]:
        if project is None and self.project is not None:
            project = self.project.id
        elif isinstance(project, str):
            if workspace is None and self.workspace is not None:
                workspace = self.workspace.id
        if project is None:
            raise MissingReferenceError("project")
        return StageApi(self._auth).list_stages(project, workspace, search, page_index, page_size)

    def delete_stage(
        self, stage: Reference, project: Optional[Reference] = None, workspace: Optional[Reference] = None
    ) -> None:
        if isinstance(stage, str):
            if project is None and self.project is not None:
                project = self.project.id
            elif isinstance(project, str):
                if workspace is None and self.workspace is not None:
                    workspace = self.workspace.id
        return StageApi(self._auth).delete_stage(stage, project, workspace)

    def create_block(
        self,
        stage: Reference,
        data: BlockInput,
        project: Optional[Reference] = None,
        workspace: Optional[Reference] = None,
    ) -> BlockOutput:
        if project is None and self.project is not None:
            project = self.project.id
        if workspace is None and self.workspace is not None:
            workspace = self.workspace.id
        if project is None:
            raise MissingReferenceError("project")
        return BlockApi(self._auth).create_block(stage, data, project, workspace)

    def update_block(
        self,
        position: int,
        stage: Reference,
        data: BlockInput,
        project: Optional[Reference] = None,
        workspace: Optional[Reference] = None,
    ) -> BlockOutput:
        if project is None and self.project is not None:
            project = self.project.id
        if workspace is None and self.workspace is not None:
            workspace = self.workspace.id
        if project is None:
            raise MissingReferenceError("project")
        return BlockApi(self._auth).update_block(position, stage, data, project, workspace)

    def delete_block(
        self,
        position: int,
        stage: Reference,
        project: Optional[Reference] = None,
        workspace: Optional[Reference] = None,
    ):
        if isinstance(stage, str):
            if project is None and self.project is not None:
                project = self.project.id
            elif isinstance(project, str):
                if workspace is None and self.workspace is not None:
                    workspace = self.workspace.id
            elif project is None and self.workspace is None:
                raise MissingReferenceError("project")
        return BlockApi(self._auth).delete_block(position, stage, project, workspace)

    def list_blocks(
        self,
        stage: Reference,
        project: Optional[Reference] = None,
        workspace: Optional[Reference] = None,
        search: Optional[str] = None,
        page_index=Page.index,
        page_size=Page.size,
    ) -> List[BlockOutput]:
        if project is None and self.project is not None:
            project = self.project.id
        if workspace is None and self.workspace is not None:
            workspace = self.workspace.id
        if project is None:
            raise MissingReferenceError("project")
        return BlockApi(self._auth).list_blocks(stage, project, workspace, search, page_index, page_size)



from enum import Enum


class ConnectionType(Enum):
    """
    Enumeration of supported connection types.
    """

    GoogleBigQuery = "GoogleBigQuery"
    """
    """

    GoogleStorage = "GoogleStorage"
    """
    """

    GitHub = "GitHub"
    """
    """

    AmazonS3 = "AmazonS3"
    """
    """

    AmazonRedShift = "AmazonRedShift"
    """
    """

    Local = "Local"
    """
    """



from __future__ import annotations
from typing import Optional, TYPE_CHECKING

if TYPE_CHECKING:
    from vectice import Reference


class Profile:
    """
    User Profile information
    """

    def __init__(self, login: str, default_workspace: Optional[Reference] = None):
        self._login = login
        self._default_workspace = default_workspace

    """
    user login/email

    """

    @property
    def login(self) -> str:
        return self._login

    """
    default workspace is no workspace is given in the API.

    """

    @property
    def default_workspace(self) -> Optional[Reference]:
        return self._default_workspace



from __future__ import annotations

import urllib
from typing import Optional, TYPE_CHECKING

from .reference import InvalidReferenceError, MissingReferenceError
from ._auth import Auth
from ._http import HttpError
from .json import ProjectOutput, PagedResponse, ProjectInput
from .http_error_handlers import HttpErrorHandler
from .workspace import WorkspaceApi

if TYPE_CHECKING:
    from vectice import Reference


class ProjectApi:
    def __init__(
        self,
        auth: Auth,
    ):
        self._auth = auth
        self._httpErrorhandler = HttpErrorHandler()

    def create_project(self, data: ProjectInput, workspace: Reference) -> ProjectOutput:
        if isinstance(workspace, int):
            url = f"/metadata/workspace/{workspace}/project"
        elif isinstance(workspace, str):
            parent_workspace = WorkspaceApi(self._auth).get_workspace(workspace)
            url = f"/metadata/workspace/{parent_workspace.id}/project"
        else:
            raise InvalidReferenceError("workspace", workspace)
        try:
            response = self._auth._post(url, data)
            return ProjectOutput(**response)
        except HttpError as e:
            raise self._httpErrorhandler.handlePostHttpError(e, "project")
        except IndexError:
            raise ValueError("The project is invalid. Please check the entered value.")

    def delete_project(self, project: Reference, workspace: Optional[Reference] = None):
        project_output = self.get_project(project, workspace)
        url = f"/metadata/project/{project_output.id}"
        try:
            response = self._auth._delete(url)
            return ProjectOutput(**response)
        except HttpError as e:
            raise self._httpErrorhandler.handleDeleteHttpError(e, "project", project)
        except IndexError:
            raise ValueError("The project is invalid. Please check the entered value.")

    def get_project(self, project: Reference, workspace: Optional[Reference] = None) -> ProjectOutput:
        if isinstance(project, int):
            url = f"/metadata/project/{project}"
        elif isinstance(project, str):
            if workspace is None:
                raise MissingReferenceError("workspace")
            parent_workspace = WorkspaceApi(self._auth).get_workspace(workspace)
            url = f"/metadata/workspace/{parent_workspace.id}/project/name/{urllib.parse.quote(project)}"
        else:
            raise InvalidReferenceError("project", project)
        try:
            response = self._auth._get(url)
            return ProjectOutput(**response)
        except HttpError as e:
            raise self._httpErrorhandler.handleGetHttpError(e, "project", project)
        except IndexError:
            raise ValueError("The project is invalid. Please check the entered value.")

    def update_project(self, data: ProjectInput, project: Reference, workspace: Reference) -> ProjectOutput:
        if isinstance(workspace, int):
            base_url = f"/metadata/workspace/{workspace}/project"
        elif isinstance(workspace, str):
            base_url = f"/metadata/workspace/name/{urllib.parse.quote(workspace)}/project"
        else:
            raise InvalidReferenceError("workspace", workspace)
        if isinstance(project, int):
            url = f"/metadata/project/{project}"
        elif isinstance(project, str):
            url = f"{base_url}/name/{urllib.parse.quote(project)}"
        else:
            raise InvalidReferenceError("project", project)
        try:
            response = self._auth._put(url, data)
            return ProjectOutput(**response)
        except HttpError as e:
            raise self._httpErrorhandler.handlePutHttpError(e, "project", project)
        except IndexError:
            raise ValueError("The project is invalid. Please check the entered value.")

    def list_projects(
        self, workspace: Reference, search: Optional[str] = None, page_index: int = 1, page_size: int = 20
    ) -> PagedResponse[ProjectOutput]:
        if isinstance(workspace, int):
            url = f"/metadata/workspace/project?index={page_index}&size={page_size}&workspaceId={workspace}"
        elif isinstance(workspace, str):
            url = f"/metadata/workspace/project?index={page_index}&size={page_size}&workspaceName={workspace}"
        else:
            raise InvalidReferenceError("workspace", workspace)
        if search:
            url = url + f"&search={search}"
        projects = self._auth._get(url)
        return PagedResponse(
            item_cls=ProjectOutput, total=projects["total"], page=projects["page"], items=projects["items"]
        )



from __future__ import annotations

import urllib
from typing import Optional, TYPE_CHECKING
from urllib.parse import urlencode

from ._auth import Auth
from ._http import HttpError
from .http_error_handlers import HttpErrorHandler
from .json import CodeInput, CodeOutput, Page, PagedResponse
from .project import ProjectApi
from .reference import InvalidReferenceError, MissingReferenceError

if TYPE_CHECKING:
    from vectice import Reference


class CodeApi:
    def __init__(self, auth: Auth):
        self._auth = auth
        self._httpErrorhandler = HttpErrorHandler()

    def create_code(
        self,
        code_data: CodeInput,
        project: Optional[Reference] = None,
        workspace: Optional[Reference] = None,
    ) -> CodeOutput:
        if code_data.name is None:
            raise ValueError('"name" must be provided in code.')
        if isinstance(project, int):
            url = f"/metadata/project/{project}/code"
        elif isinstance(project, str):
            parent_project = ProjectApi(self._auth).get_project(project, workspace)
            url = f"/metadata/project/{parent_project.id}/code"
        else:
            raise InvalidReferenceError("project", project)
        try:
            response = self._auth._post(url, code_data)
            return CodeOutput(**response)
        except HttpError as e:
            raise self._httpErrorhandler.handlePostHttpError(e, "code")

    def get_code(
        self,
        code: Reference,
        project: Optional[Reference] = None,
        workspace: Optional[Reference] = None,
    ) -> CodeOutput:
        if not isinstance(code, int) and not isinstance(code, str):
            raise InvalidReferenceError("code", code)
        if project is None:
            raise MissingReferenceError("code", "project")
        parent_project = ProjectApi(self._auth).get_project(project, workspace)
        if isinstance(code, int):
            url = f"/metadata/project/{parent_project.id}/code/{code}"
        else:
            url = f"/metadata/project/{parent_project.id}/code/name/{urllib.parse.quote(code)}"
        try:
            response = self._auth._get(url)
            return CodeOutput(response)
        except HttpError as e:
            raise self._httpErrorhandler.handleGetHttpError(e, "code", code)
        except IndexError:
            raise ValueError("The code is invalid. Please check the entered value.")

    # There is No route in the backend to list code
    # Waiting for BE implementation : https://app.shortcut.com/vectice/story/23415/list-code-rest-api
    def list_codes(
        self,
        project: Reference,
        workspace: Optional[Reference] = None,
        search: Optional[str] = None,
        page_index=Page.index,
        page_size=Page.size,
    ) -> PagedResponse[CodeOutput]:
        if isinstance(project, int):
            url = f"/metadata/project/{project}/code"
        elif isinstance(project, str):
            parent_project = ProjectApi(self._auth).get_project(project, workspace)
            url = f"/metadata/project/{parent_project.id}/code"
        else:
            raise InvalidReferenceError("project", project)
        try:
            queries = {"index": page_index, "size": page_size}
            if search:
                queries["search"] = search
            response = self._auth._get(f"{url}?{urlencode(queries)}")
            result = PagedResponse(
                item_cls=CodeOutput, total=response["total"], page=response["page"], items=response["items"]
            )
            return result
        except HttpError as e:
            raise self._httpErrorhandler.handleGetHttpError(e, "project", project)

    def update_code(
        self,
        code_data: CodeOutput,
        project: Optional[Reference] = None,
        workspace: Optional[Reference] = None,
    ) -> CodeOutput:
        code_output = self.get_code(code_data.id, project, workspace)
        url = f"/metadata/project/{code_output.project_id}/code/{code_data.id}"
        try:
            response = self._auth._put(url, code_data)
            return CodeOutput(**response)
        except HttpError as e:
            raise self._httpErrorhandler.handlePutHttpError(e, "code", code_data.id)

    def delete_code(
        self,
        code: Reference,
        project: Optional[Reference] = None,
        workspace: Optional[Reference] = None,
    ) -> None:
        try:
            code_object = self.get_code(code, project, workspace)
            if isinstance(code, str):
                self._auth._delete(f"/metadata/project/{code_object.project_id}/code/{code_object.id}")
            else:
                self._auth._delete(f"/metadata/project/{code_object.project_id}/code/{code}")
        except HttpError as e:
            raise self._httpErrorhandler.handleGetHttpError(e, "code", code)
